#! /usr/env/python
"""
Python implementation of ModelGrid, a base class used to create and manage
grids for 2D numerical models.

Getting Information about a Grid
--------------------------------
The following attributes, properties, and methods provide data about the grid,
its geometry, and the connectivity among the various elements. Each grid
element has an ID number, which is also its position in an array that
contains information about that type of element. For example, the *x*
coordinate of node 5 would be found at `grid.node_x[5]`.

The naming of grid-element arrays is *attribute*`_at_`*element*, where
*attribute* is the name of the data in question, and *element* is the element
to which the attribute applies. For example, the property `node_at_cell`
contains the ID of the node associated with each cell. For example,
`node_at_cell[3]` contains the *node ID* of the node associated with cell 3.
The *attribute* is singular if there is only one value per element; for
example, there is only one node associated with each cell. It is plural when
there are multiple values per element; for example, the `faces_at_cell` array
contains multiple faces for each cell. Exceptions to these general rules are
functions that return indices of a subset of all elements of a particular type.
For example, you can obtain an array with IDs of only the core nodes using
`core_nodes`, while `active_links` provides an array of IDs of active links
(only). Finally, attributes that represent a measurement of something, such as
the length of a link or the surface area of a cell, are described using `_of_`,
as in the example `area_of_cell`.

Information about the grid as a whole
+++++++++++++++++++++++++++++++++++++

.. autosummary::
    :toctree: generated/

    ~landlab.grid.radial.RadialModelGrid.axis_name
    ~landlab.grid.radial.RadialModelGrid.axis_units
    ~landlab.grid.radial.RadialModelGrid.from_dict
    ~landlab.grid.radial.RadialModelGrid.move_origin
    ~landlab.grid.radial.RadialModelGrid.node_axis_coordinates
    ~landlab.grid.radial.RadialModelGrid.number_of_active_faces
    ~landlab.grid.radial.RadialModelGrid.number_of_active_links
    ~landlab.grid.radial.RadialModelGrid.number_of_cells
    ~landlab.grid.radial.RadialModelGrid.number_of_core_cells
    ~landlab.grid.radial.RadialModelGrid.number_of_core_nodes
    ~landlab.grid.radial.RadialModelGrid.number_of_elements
    ~landlab.grid.radial.RadialModelGrid.number_of_faces
    ~landlab.grid.radial.RadialModelGrid.number_of_fixed_links
    ~landlab.grid.radial.RadialModelGrid.number_of_links
    ~landlab.grid.radial.RadialModelGrid.number_of_nodes
    ~landlab.grid.radial.RadialModelGrid.number_of_nodes_in_shell
    ~landlab.grid.radial.RadialModelGrid.number_of_patches
    ~landlab.grid.radial.RadialModelGrid.number_of_shells
    ~landlab.grid.radial.RadialModelGrid.save
    ~landlab.grid.radial.RadialModelGrid.spacing_of_shells


Information about nodes
+++++++++++++++++++++++

.. autosummary::
    :toctree: generated/

    ~landlab.grid.radial.RadialModelGrid.active_link_dirs_at_node
    ~landlab.grid.radial.RadialModelGrid.all_node_azimuths_map
    ~landlab.grid.radial.RadialModelGrid.all_node_distances_map
    ~landlab.grid.radial.RadialModelGrid.boundary_nodes
    ~landlab.grid.radial.RadialModelGrid.cell_area_at_node
    ~landlab.grid.radial.RadialModelGrid.cell_at_node
    ~landlab.grid.radial.RadialModelGrid.closed_boundary_nodes
    ~landlab.grid.radial.RadialModelGrid.core_nodes
    ~landlab.grid.radial.RadialModelGrid.downwind_links_at_node
    ~landlab.grid.radial.RadialModelGrid.fixed_gradient_boundary_nodes
    ~landlab.grid.radial.RadialModelGrid.fixed_value_boundary_nodes
    ~landlab.grid.radial.RadialModelGrid.link_at_node_is_downwind
    ~landlab.grid.radial.RadialModelGrid.link_at_node_is_upwind
    ~landlab.grid.radial.RadialModelGrid.link_dirs_at_node
    ~landlab.grid.radial.RadialModelGrid.links_at_node
    ~landlab.grid.radial.RadialModelGrid.neighbors_at_node
    ~landlab.grid.radial.RadialModelGrid.node_at_cell
    ~landlab.grid.radial.RadialModelGrid.node_at_core_cell
    ~landlab.grid.radial.RadialModelGrid.node_at_link_head
    ~landlab.grid.radial.RadialModelGrid.node_at_link_tail
    ~landlab.grid.radial.RadialModelGrid.node_axis_coordinates
    ~landlab.grid.radial.RadialModelGrid.node_is_boundary
    ~landlab.grid.radial.RadialModelGrid.node_x
    ~landlab.grid.radial.RadialModelGrid.node_y
    ~landlab.grid.radial.RadialModelGrid.nodes
    ~landlab.grid.radial.RadialModelGrid.nodes_at_patch
    ~landlab.grid.radial.RadialModelGrid.number_of_core_nodes
    ~landlab.grid.radial.RadialModelGrid.number_of_links_at_node
    ~landlab.grid.radial.RadialModelGrid.number_of_nodes
    ~landlab.grid.radial.RadialModelGrid.number_of_nodes_in_shell
    ~landlab.grid.radial.RadialModelGrid.open_boundary_nodes
    ~landlab.grid.radial.RadialModelGrid.patches_at_node
    ~landlab.grid.radial.RadialModelGrid.radius_at_node
    ~landlab.grid.radial.RadialModelGrid.set_nodata_nodes_to_closed
    ~landlab.grid.radial.RadialModelGrid.set_nodata_nodes_to_fixed_gradient
    ~landlab.grid.radial.RadialModelGrid.status_at_node
    ~landlab.grid.radial.RadialModelGrid.unit_vector_sum_xcomponent_at_node
    ~landlab.grid.radial.RadialModelGrid.unit_vector_sum_ycomponent_at_node
    ~landlab.grid.radial.RadialModelGrid.upwind_links_at_node
    ~landlab.grid.radial.RadialModelGrid.x_of_node
    ~landlab.grid.radial.RadialModelGrid.y_of_node


Information about links
+++++++++++++++++++++++

.. autosummary::
    :toctree: generated/

    ~landlab.grid.radial.RadialModelGrid.active_link_dirs_at_node
    ~landlab.grid.radial.RadialModelGrid.active_links
    ~landlab.grid.radial.RadialModelGrid.angle_of_link
    ~landlab.grid.radial.RadialModelGrid.downwind_links_at_node
    ~landlab.grid.radial.RadialModelGrid.face_at_link
    ~landlab.grid.radial.RadialModelGrid.fixed_links
    ~landlab.grid.radial.RadialModelGrid.length_of_link
    ~landlab.grid.radial.RadialModelGrid.link_at_face
    ~landlab.grid.radial.RadialModelGrid.link_at_node_is_downwind
    ~landlab.grid.radial.RadialModelGrid.link_at_node_is_upwind
    ~landlab.grid.radial.RadialModelGrid.link_dirs_at_node
    ~landlab.grid.radial.RadialModelGrid.links_at_node
    ~landlab.grid.radial.RadialModelGrid.node_at_link_head
    ~landlab.grid.radial.RadialModelGrid.node_at_link_tail
    ~landlab.grid.radial.RadialModelGrid.number_of_active_links
    ~landlab.grid.radial.RadialModelGrid.number_of_fixed_links
    ~landlab.grid.radial.RadialModelGrid.number_of_links
    ~landlab.grid.radial.RadialModelGrid.number_of_links_at_node
    ~landlab.grid.radial.RadialModelGrid.resolve_values_on_active_links
    ~landlab.grid.radial.RadialModelGrid.resolve_values_on_links
    ~landlab.grid.radial.RadialModelGrid.status_at_link
    ~landlab.grid.radial.RadialModelGrid.unit_vector_xcomponent_at_link
    ~landlab.grid.radial.RadialModelGrid.unit_vector_ycomponent_at_link
    ~landlab.grid.radial.RadialModelGrid.upwind_links_at_node

Information about cells
+++++++++++++++++++++++

.. autosummary::
    :toctree: generated/

    ~landlab.grid.radial.RadialModelGrid.area_of_cell
    ~landlab.grid.radial.RadialModelGrid.cell_area_at_node
    ~landlab.grid.radial.RadialModelGrid.cell_at_node
    ~landlab.grid.radial.RadialModelGrid.core_cells
    ~landlab.grid.radial.RadialModelGrid.faces_at_cell
    ~landlab.grid.radial.RadialModelGrid.node_at_cell
    ~landlab.grid.radial.RadialModelGrid.node_at_core_cell
    ~landlab.grid.radial.RadialModelGrid.number_of_cells
    ~landlab.grid.radial.RadialModelGrid.number_of_core_cells
    ~landlab.grid.radial.RadialModelGrid.number_of_faces_at_cell

Information about faces
+++++++++++++++++++++++

.. autosummary::
    :toctree: generated/

    ~landlab.grid.radial.RadialModelGrid.active_faces
    ~landlab.grid.radial.RadialModelGrid.face_at_link
    ~landlab.grid.radial.RadialModelGrid.faces_at_cell
    ~landlab.grid.radial.RadialModelGrid.link_at_face
    ~landlab.grid.radial.RadialModelGrid.number_of_active_faces
    ~landlab.grid.radial.RadialModelGrid.number_of_faces
    ~landlab.grid.radial.RadialModelGrid.number_of_faces_at_cell
    ~landlab.grid.radial.RadialModelGrid.width_of_face

Information about patches
+++++++++++++++++++++++++

.. autosummary::
    :toctree: generated/

    ~landlab.grid.radial.RadialModelGrid.nodes_at_patch
    ~landlab.grid.radial.RadialModelGrid.number_of_patches
    ~landlab.grid.radial.RadialModelGrid.patches_at_node

Data Fields in ModelGrid
------------------------
:class:`~.ModelGrid` inherits from the :class:`~.ModelDataFields` class. This
provides `~.ModelGrid`, and its subclasses, with the ability to, optionally,
store data values that are associated with the different types grid elements
(nodes, cells, etc.). In particular, as part of ``ModelGrid.__init__()``,
data field *groups* are added to the `ModelGrid` that provide containers to
put data fields into. There is one group for each of the eight grid elements
(node, cell, link, face, core_node, core_cell, active_link, and active_face).

To access these groups, use the same methods as accessing groups with
`~.ModelDataFields`. ``ModelGrid.__init__()`` adds the following attributes to
itself that provide access to the values groups:

.. autosummary::
    :toctree: generated/
    :nosignatures:

    ~landlab.grid.radial.RadialModelGrid.at_node
    ~landlab.grid.radial.RadialModelGrid.at_cell
    ~landlab.grid.radial.RadialModelGrid.at_link
    ~landlab.grid.radial.RadialModelGrid.at_face

Each of these attributes returns a ``dict``-like object whose keys are value
names as strings and values are numpy arrays that gives quantities at
grid elements.


Create Field Arrays
+++++++++++++++++++
:class:`~.ModelGrid` inherits several useful methods for creating new data
fields and adding new data fields to a ModelGrid instance. Methods to add or
create a new data array follow the ``numpy`` syntax for creating arrays. The
folowing methods create and, optionally, initialize new arrays. These arrays
are of the correct size but a new field will not be added to the field:

.. autosummary::
    :toctree: generated/
    :nosignatures:

    ~landlab.field.grouped.ModelDataFields.empty
    ~landlab.field.grouped.ModelDataFields.ones
    ~landlab.field.grouped.ModelDataFields.zeros

Add Fields to a ModelGrid
+++++++++++++++++++++++++
Unlike with the equivalent numpy functions, these do not take a size argument
as the size of the returned arrays is determined from the size of the
ModelGrid. However, the keyword arguments are the same as those of the numpy
equivalents.

The following methods will create a new array and add a reference to that
array to the ModelGrid:

.. autosummary::
    :toctree: generated/
    :nosignatures:

    ~landlab.grid.radial.RadialModelGrid.add_empty
    ~landlab.grid.radial.RadialModelGrid.add_field
    ~landlab.grid.radial.RadialModelGrid.add_ones
    ~landlab.grid.radial.RadialModelGrid.add_zeros
    ~landlab.grid.radial.RadialModelGrid.delete_field
    ~landlab.grid.radial.RadialModelGrid.set_units

These methods operate in the same way as the previous set except that, in
addition to creating a new array, the newly-created array is added to the
ModelGrid. The calling signature is the same but with the addition of an
argument that gives the name of the new field as a string. The additional
method, :meth:`~.ModelDataFields.add_field`, adds a previously allocation
array to the ModelGrid. If the array is of the incorrect size it will raise
``ValueError``.

Query Fields
++++++++++++
Use the following methods/attributes get information about the stored data
fields:

.. autosummary::
    :toctree: generated/
    :nosignatures:

    ~landlab.field.grouped.ModelDataFields.size
    ~landlab.field.grouped.ModelDataFields.keys
    ~landlab.field.grouped.ModelDataFields.has_group
    ~landlab.field.grouped.ModelDataFields.has_field
    ~landlab.grid.radial.RadialModelGrid.field_units
    ~landlab.grid.radial.RadialModelGrid.field_values
    ~landlab.field.grouped.ModelDataFields.groups

    i.e., call, e.g. mg.has_field('node', 'my_field_name')

    # START HERE check that all functions listed below are included above,
    # ignore ones that start with underscores(_)

Gradients, fluxes, and divergences on the grid
----------------------------------------------

Landlab is designed to easily calculate gradients in quantities across the
grid, and to construct fluxes and flux divergences from them. Because these
calculations tend to be a little more involved than property lookups, the
methods tend to start with `calc_`.

.. autosummary::
    :toctree: generated/

    ~landlab.grid.radial.RadialModelGrid.calc_diff_at_link
    ~landlab.grid.radial.RadialModelGrid.calc_flux_div_at_node
    ~landlab.grid.radial.RadialModelGrid.calc_grad_at_link
    ~landlab.grid.radial.RadialModelGrid.calc_grad_at_patch
    ~landlab.grid.radial.RadialModelGrid.calc_net_flux_at_node
    ~landlab.grid.radial.RadialModelGrid.calc_slope_at_node
    ~landlab.grid.radial.RadialModelGrid.calc_slope_at_patch
    ~landlab.grid.radial.RadialModelGrid.calc_unit_normal_at_patch

Mappers
-------

These methods allow mapping of values defined on one grid element type onto a
second, e.g., mapping upwind node values onto links, or mean link values onto
nodes.

.. autosummary::
    :toctree: generated/

    ~landlab.grid.radial.RadialModelGrid.map_downwind_node_link_max_to_node
    ~landlab.grid.radial.RadialModelGrid.map_downwind_node_link_mean_to_node
    ~landlab.grid.radial.RadialModelGrid.map_link_head_node_to_link
    ~landlab.grid.radial.RadialModelGrid.map_link_tail_node_to_link
    ~landlab.grid.radial.RadialModelGrid.map_link_vector_to_nodes
    ~landlab.grid.radial.RadialModelGrid.map_max_of_link_nodes_to_link
    ~landlab.grid.radial.RadialModelGrid.map_max_of_node_links_to_node
    ~landlab.grid.radial.RadialModelGrid.map_mean_of_link_nodes_to_link
    ~landlab.grid.radial.RadialModelGrid.map_min_of_link_nodes_to_link
    ~landlab.grid.radial.RadialModelGrid.map_min_of_node_links_to_node
    ~landlab.grid.radial.RadialModelGrid.map_node_to_cell
    ~landlab.grid.radial.RadialModelGrid.map_upwind_node_link_max_to_node
    ~landlab.grid.radial.RadialModelGrid.map_upwind_node_link_mean_to_node
    ~landlab.grid.radial.RadialModelGrid.map_value_at_downwind_node_link_max_to_node
    ~landlab.grid.radial.RadialModelGrid.map_value_at_max_node_to_link
    ~landlab.grid.radial.RadialModelGrid.map_value_at_min_node_to_link
    ~landlab.grid.radial.RadialModelGrid.map_value_at_upwind_node_link_max_to_node

Boundary condition control
--------------------------

These are the primary properties for getting and setting the grid boundary
conditions. Changes made to :meth:`~.ModelGrid.status_at_node` and
:meth:`~.ModelGrid.status_at_node` will automatically update the conditions
defined at other grid elements automatically.

.. autosummary::
    :toctree: generated/

   .....:     
    ~landlab.grid.radial.RadialModelGrid.active_faces
    ~landlab.grid.radial.RadialModelGrid.active_links
    ~landlab.grid.radial.RadialModelGrid.boundary_nodes
    ~landlab.grid.radial.RadialModelGrid.closed_boundary_nodes
    ~landlab.grid.radial.RadialModelGrid.core_cells
    ~landlab.grid.radial.RadialModelGrid.core_nodes
    ~landlab.grid.radial.RadialModelGrid.fixed_gradient_boundary_nodes
    ~landlab.grid.radial.RadialModelGrid.fixed_links
    ~landlab.grid.radial.RadialModelGrid.fixed_value_boundary_nodes
    ~landlab.grid.radial.RadialModelGrid.node_at_core_cell
    ~landlab.grid.radial.RadialModelGrid.node_is_boundary
    ~landlab.grid.radial.RadialModelGrid.number_of_active_faces
    ~landlab.grid.radial.RadialModelGrid.number_of_active_links
    ~landlab.grid.radial.RadialModelGrid.number_of_core_cells
    ~landlab.grid.radial.RadialModelGrid.number_of_core_nodes
    ~landlab.grid.radial.RadialModelGrid.number_of_fixed_links
    ~landlab.grid.radial.RadialModelGrid.open_boundary_nodes
    ~landlab.grid.radial.RadialModelGrid.set_nodata_nodes_to_closed
    ~landlab.grid.radial.RadialModelGrid.set_nodata_nodes_to_fixed_gradient
    ~landlab.grid.radial.RadialModelGrid.status_at_link
    ~landlab.grid.radial.RadialModelGrid.status_at_node

Identifying node subsets
------------------------

These methods are useful in identifying subsets of nodes, e.g., closest node
to a point; nodes at edges.

(None are available for this grid type)

Surface analysis
----------------

These methods permit the kinds of surface analysis that you might expect to
find in GIS software.

.. autosummary::
    :toctree: generated/

    ~landlab.grid.radial.RadialModelGrid.calc_aspect_at_node
    ~landlab.grid.radial.RadialModelGrid.calc_hillshade_at_node
    ~landlab.grid.radial.RadialModelGrid.calc_slope_at_node

Notes
-----
It is important that when creating a new grid class that inherits from
``ModelGrid``, to call ``ModelGrid.__init__()`` in the new grid's
``__init__()``. For example, the new class's __init__ should contain the
following code,

.. code-block:: python

    class NewGrid(ModelGrid):
        def __init__(self, *args, **kwds):
            ModelGrid.__init__(self, **kwds)
            # Code that initializes the NewGrid

Without this, the new grid class will not have the ``at_*`` attributes.
"""

import numpy
from six.moves import range

from .voronoi import VoronoiDelaunayGrid
from landlab.utils.decorators import deprecated


class RadialModelGrid(VoronoiDelaunayGrid):

    """Grid of concentric circles.

    This inherited class implements a circular grid in which grid nodes are
    placed at regular radial and semi-regular arc-wise intervals. That is,
    if the radial spacing between *shells* is *dr*, the nodes are placed around
    the circular shell at regular intervals that get as close as possible to
    *dr*. The points are then arranged in a Delaunay triangulation with Voronoi
    cells. Within each ring, nodes are numbered according to Landlab
    convention, from the first node counterclockwise of east. Numbering
    begins at the centermost node and works outwards through the rings.

    Parameters
    ----------
    num_shells : int
        Number of rings in the grid.
    dr : float, optional
        Radial interval for rings.
    origin_x : float, optional
        x-coordinate of origin node.
    origin_y : float, optional
        y-coordinate of origin node.

    Returns
    -------
    RadialModelGrid
        A newly-created grid.

    Examples
    --------
    A grid with just one ring will have a node at the origin surrounded
    by six other nodes.

    >>> from landlab import RadialModelGrid
    >>> omg = RadialModelGrid(num_shells=1, dr=1., origin_x=0., origin_y=0.)
    >>> omg.number_of_nodes
    7
    >>> omg.number_of_cells
    1

    A second rings will have 13 nodes.

    >>> omg = RadialModelGrid(2)
    >>> omg.number_of_nodes
    20
    """

    def __init__(self, num_shells=0, dr=1.0, origin_x=0.0, origin_y=0.0,
                 **kwds):
        """Create a circular grid.

        Create a circular grid in which grid nodes are placed at regular
        radial and semi-regular arc-wise intervals. That is, if the radial
        spacing between *shells* is *dr*, the nodes are placed around the
        circular shell at regular intervals that get as close as possible to
        *dr*.  The points are then arranged in a Delaunay triangulation with
        Voronoi cells.

        Parameters
        ----------
        num_shells : int
            Number of rings in the grid.
        dr : float, optional
            Radial interval for rings.
        origin_x : float, optional
            x-coordinate of origin node.
        origin_y : float, optional
            y-coordinate of origin node.

        Returns
        -------
        RadialModelGrid
            A newly-created grid.

        Examples
        --------
        A grid with just one ring will have a node at the origin surrounded
        by six other nodes.

        >>> from landlab import RadialModelGrid
        >>> omg = RadialModelGrid(num_shells=1, dr=1., origin_x=0.,
        ...                       origin_y=0.)
        >>> omg.number_of_nodes
        7
        >>> omg.number_of_cells
        1

        A second rings will have 13 nodes.

        >>> omg = RadialModelGrid(2)
        >>> omg.number_of_nodes
        20
        """
        # Set number of nodes, and initialize if caller has given dimensions
        self._origin_x = origin_x
        self._origin_y = origin_y
        if num_shells > 0:
            self._initialize(num_shells, dr, origin_x, origin_y)
        super(RadialModelGrid, self).__init__(**kwds)

    @classmethod
    def from_dict(cls, params):
        num_shells = params['num_shells']
        dr = params.get('dr', 1.)
        origin = params.get('origin', (0., 0.))

        return cls(num_shells=num_shells, dr=dr, origin_x=origin[0],
                   origin_y=origin[1])

    def _initialize(self, num_shells, dr, origin_x=0.0, origin_y=0.0):
        [pts, npts] = self._create_radial_points(num_shells, dr)
        self._n_shells = int(num_shells)
        self._dr = dr
        super(RadialModelGrid, self)._initialize(pts[:, 0], pts[:, 1])

    def _create_radial_points(self, num_shells, dr, origin_x=0.0,
                              origin_y=0.0):
        """Create a set of points on concentric circles.

        Creates and returns a set of (x,y) points placed in a series of
        concentric circles around the origin.
        """
        shells = numpy.arange(0, num_shells) + 1
        twopi = 2 * numpy.pi
        # number of points in each shell
        n_pts_in_shell = numpy.round(twopi * shells)
        dtheta = twopi / n_pts_in_shell
        npts = int(sum(n_pts_in_shell) + 1)
        pts = numpy.zeros((npts, 2))
        r = shells * dr
        startpt = 1
        for i in numpy.arange(0, num_shells):
            theta = (dtheta[i] * numpy.arange(0, n_pts_in_shell[i]) +
                     dtheta[i] / (i + 1))
            ycoord = r[i] * numpy.sin(theta)
            if numpy.isclose(ycoord[-1], 0.):
                # this modification necessary to force the first ring to
                # follow our new CCW from E numbering convention (DEJH, Nov15)
                ycoord[-1] = 0.
                pts[startpt:(startpt + int(n_pts_in_shell[i])),
                    0] = numpy.roll(r[i] * numpy.cos(theta), 1)
                pts[startpt:(startpt + int(n_pts_in_shell[i])),
                    1] = numpy.roll(ycoord, 1)
            else:
                pts[startpt:(startpt + int(n_pts_in_shell[i])),
                    0] = r[i] * numpy.cos(theta)
                pts[startpt:(startpt + int(n_pts_in_shell[i])),
                    1] = ycoord
            startpt += int(n_pts_in_shell[i])
        pts[:, 0] += origin_x
        pts[:, 1] += origin_y

        return pts, npts

    @property
    def number_of_shells(self):
        """Number of node shells in grid.

        Returns
        -------
        int
            The number of node shells in the radial grid (not counting the
            center node).
        """
        return self._n_shells

    @property
    @deprecated(use='spacing_of_shells', version=1.0)
    def shell_spacing(self):
        """Fixed distance between shells."""
        return self._dr

    @property
    def spacing_of_shells(self):
        """Fixed distance between shells."""
        return self._dr

    @property
    def number_of_nodes_in_shell(self):
        """Number of nodes in each shell.

        Returns
        -------
        int
            Number of nodes in each shell, excluding the center node.
        """
        try:
            return self._nnodes_inshell
        except AttributeError:
            n_pts_in_shell = numpy.round(2. * numpy.pi * (
                numpy.arange(self.number_of_shells, dtype=float) + 1.))
            self._nnodes_inshell = n_pts_in_shell.astype(int)
            return self._nnodes_inshell

    @property
    def radius_at_node(self):
        """Distance for center node to each node.

        Returns
        -------
        ndarray of float
            The distance from the center node of each node.

        >>> mg = RadialModelGrid(num_shells=2)
        >>> mg.radius_at_node
        array([ 2.,  2.,  2.,  2.,  2.,  1.,  1.,  2.,  0.,  1.,  1.,  2.,  2.,
                1.,  1.,  2.,  2.,  2.,  2.,  2.])
        """
        try:
            return self._node_radii
        except AttributeError:
            self._node_radii = numpy.sqrt(numpy.square(self.node_x -
                                                       self._origin_x) +
                                          numpy.square(self.node_y -
                                                       self._origin_y))
            return self._node_radii



#! /usr/bin/env python

import numpy as np
from scipy import ndimage

from landlab import Implements
from landlab.framework.interfaces import (BmiBase, BmiUniformRectilinear,
                                          BadVarNameError, GRID_TYPE_UNIFORM)


#@ImplementsOrRaise(BmiBase, BmiUniformRectilinear)
@Implements(BmiBase, BmiUniformRectilinear)
class Diff2D(object): #pylint:disable-msg=R0902,R0904
    """
    Basic Modeling Interface implementation for a 2d diffusion process.
    """

    __implements__ = (BmiBase, BmiUniformRectilinear, )

    var_units = {'surface__elevation': 'meters'}
    name = '2D Diffusion'

    _input_var_names = [
        'surface__elevation',
    ]

    _output_var_names = [
        'surface__elevation',
    ]

    def __init__(self):
        self._time_step = 0.
        self._shape = (0, 0)
        self._spacing = (0., 0.)
        self._origin = (0., 0.)
        self._start_time = 0.
        self._end_time = np.finfo ('d').max
        self._time = self._start_time

        self._elevation = None
        self._temp_elevation = None

        self._var_names = set(self.output_vars + self.input_vars)
        self._value = {}

        self._stencil = np.array((3, 3), dtype=np.float64)

    @staticmethod
    def _set_bc(data):
        """
        Set boundary conditions for 2D data.

        Values on left, right, and top edge are 0. Values on the bottom edge
        are a parabola centered on the edge with roots on the left and right
        boundaries.

        :data: 2D numpy array
        """
        data[:, 0] = 0.
        data[:, -1] = 0.
        data[0, :] = 0.

        top_column = data.shape[1] - 1
        data[-1, :] = .25*top_column**2 - (np.arange(top_column + 1) -
                                           .5*top_column)**2

    def initialize(self, file_name=None):
        """
        Initialve the model from a file, or use default values. The
        initialization file is one line with values,

        ${TIME_STEP}, ${NUMBER_OF_ROWS}, ${NUMBER_OF_COLUMNS}

        :keyword file_name: Name of initialization file, or None for defaults
        """
        if file_name is not None:
            with open(file_name, 'r') as init_file:
                vals = init_file.readline().split(',')
            self._time_step, row_count, column_count = (
                float(vals[0]),
                int(vals[1]),
                int(vals[2]),
            )
            self._shape = (row_count, column_count)
        else:
            self._time_step = 1.
            self._shape = (20, 10)

        self._spacing = (1., 1.)
        self._origin = (0., 0.)

        self._elevation = np.zeros(self._shape)
        self._set_bc(self._elevation)
        self._temp_elevation = self._elevation.copy()

        self._value['surface__elevation'] = self._elevation

        row_spacing, column_spacing = self._spacing
        self._stencil = np.array(
            [[0., column_spacing**2, 0.],
             [row_spacing**2, 0., row_spacing**2],
             [0., column_spacing**2, 0.]]
        ) / (2. * (row_spacing**2 + column_spacing**2)) * self._time_step

    def update(self):
        """
        Update the model for a single time step.
        """
        ndimage.convolve(self._elevation, self._stencil,
                         output=self._temp_elevation)
        self._set_bc(self._temp_elevation)
        self._elevation[:] = self._temp_elevation

        self._time += self._time_step

    def update_until(self, stop_time):
        """
        Update the model until a given time.

        :stop_time: Update model until this time
        """
        while 1:
            next_time = self.get_current_time() + self._time_step
            if next_time > stop_time:
                break
            self.update()

        partial_step = stop_time - self.get_current_time()
        if partial_step > 0:
            self._stencil *= partial_step / self._time_step
            self.update()
            self._stencil *= self._time_step / partial_step

    def finalize(self):
        """
        Reset model values, clean up resources used.
        """
        self._time_step = 0.
        self._time = 0.

        self._elevation = np.array([])
        self._temp_elevation = np.array([])

    def get_var_type(self, long_var_name):
        """
        Type of a model variable.

        :long_var_name: Standard name string

        :return: Type as a numpy.dtype string
        """
        return str (self.get_value (long_var_name).dtype)

    def get_var_units(self, long_var_name):
        """
        Units of a model variable.

        :long_var_name: Standard name string

        :returns: Units as a udunits string
        """
        return self.var_units[long_var_name]

    def get_var_rank(self, long_var_name):
        """
        Rank of a model variable.

        :long_var_name: Standard name string

        :returns: Number of dimension
        """
        if long_var_name in self._var_names:
            return len(self._spacing)
        else:
            raise BadVarNameError(long_var_name)

    def get_value(self, long_var_name):
        """
        Values of a model variable.

        :long_var_name: Standard name string

        :returns: Numpy ND array of values
        """
        return self._value[long_var_name]

    def get_value_at_indices(self, long_var_name, indices):
        """
        Values of a model variable.

        :long_var_name: Standard name string
        :indices: Indices of (flattened) value array

        :returns: Numpy ND array of values
        """
        return self.get_value(long_var_name)[indices]

    def set_value(self, long_var_name, src):
        """
        Set values of a model variable.

        :long_var_name: Standard name string
        :src: Array of values to set
        """
        val = self.get_value(long_var_name)
        val[:] = src

    def set_value_at_indices(self, long_var_name, indices, src):
        """
        Set values of a model variable at given indices.

        :long_var_name: Standard name string
        :indices: Indices of (flattened) value array
        :src: Array of values to set
        """
        val = self.get_value(long_var_name)
        val[indices] = src

    def get_component_name(self):
        """
        Name of the component

        :returns: Name of the component as a string
        """
        return self.name

    def get_input_var_names(self):
        """
        List of input var names as standard name strings

        :returns: List of standard names
        """
        return self.input_vars

    def get_output_var_names(self):
        """
        List of output var names as standard name strings

        :returns: List of standard names
        """
        return self.output_vars

    def get_grid_shape(self, long_var_name):
        """
        Shape of the grid on which long_var_name is defined.

        :long_var_name: Standard name string

        :returns: Shape of the variable as a tuple
        """
        return self.get_value(long_var_name).shape

    def get_grid_spacing(self, long_var_name):
        """
        Row and column spacing of the grid on which long_var_name is defined.

        :long_var_name: Standard name string

        :returns: Spacing of the variable as a tuple
        """
        if long_var_name in self._var_names:
            return self._spacing
        else:
            raise BadVarNameError(long_var_name)

    def get_grid_origin(self, long_var_name):
        """
        Origin of the grid on which long_var_name is defined.

        :long_var_name: Standard name string

        :returns: Origin coordinates of the variable as a tuple
        """
        if long_var_name in self._var_names:
            return self._origin
        else:
            raise BadVarNameError(long_var_name)

    def get_grid_type(self, long_var_name):
        """
        Grid-type on which long_var_name is defined.

        :long_var_name: Standard name string

        :returns: Grid-type as a BmiGridType
        """
        if long_var_name in self._var_names:
            return GRID_TYPE_UNIFORM
        else:
            raise BadVarNameError(long_var_name)

    def get_time_step(self):
        """
        Model time step.

        :returns: Time step
        """
        return self._time_step

    def get_start_time(self):
        """
        Model start time.

        :returns: Start time
        """
        return self._start_time

    def get_end_time(self):
        """
        Model end time.

        :returns: End time
        """
        return self._end_time

    def get_current_time(self):
        """
        Current time of the model.

        :returns: Current model time
        """
        return self._time

def main ():
    """
    Run some unit tests
    """
    import doctest
    doctest.testmod ()

if __name__ == '__main__':
    main ()



#! /usr/bin/env python

from nose.tools import assert_true, assert_false, assert_raises
try:
    from nose.tools import assert_is_not, assert_is, assert_set_equal, assert_dict_equal
except ImportError:
    from landlab.testing.tools import (assert_is_not, assert_is, assert_set_equal,
                                       assert_dict_equal)
import numpy as np
from numpy.testing import assert_array_equal

from landlab.field import ModelDataFields, GroupError, FieldError


def test_init():
    fields = ModelDataFields()
    assert_set_equal(set(), fields.groups)


def test_new_field_location():
    fields = ModelDataFields()
    fields.new_field_location('node', 12)
    assert_set_equal(set(['node']), fields.groups)


def test_add_existing_group():
    fields = ModelDataFields()
    fields.new_field_location('node', size=12)
    assert_raises(ValueError, fields.new_field_location, 'node', size=24)


def test_add_multiple_groups():
    fields = ModelDataFields()
    fields.new_field_location('node', 12)
    fields.new_field_location('cell', 2)
    fields.new_field_location('face', 7)
    fields.new_field_location('link', 7)
    assert_set_equal(set(['node', 'cell', 'face', 'link']), fields.groups)


def test_ones():
    fields = ModelDataFields()
    fields.new_field_location('node', 12)
    fields.new_field_location('cell', 2)

    value_array = fields.ones('node')
    assert_array_equal(np.ones(12), value_array)

    value_array = fields.ones('cell')
    assert_array_equal(np.ones(2), value_array)


def test_add_ones():
    fields = ModelDataFields()
    fields.new_field_location('node', 12)
    fields.new_field_location('cell', 2)

    fields.add_ones('z', at='node')
    assert_array_equal(np.ones(12), fields['node']['z'])
    assert_array_equal(np.ones(12), fields.field_values('node', 'z'))

    fields.add_ones('z', at='cell')
    assert_array_equal(np.ones(2), fields['cell']['z'])
    assert_array_equal(np.ones(2), fields.field_values('cell', 'z'))


def test_add_ones_return_value():
    fields = ModelDataFields()
    fields.new_field_location('node', 12)
    fields.new_field_location('cell', 2)

    rtn_value = fields.add_ones('z', at='node')
    assert_array_equal(rtn_value, np.ones(12))
    assert_is(rtn_value, fields['node']['z'])
    assert_is(rtn_value, fields.field_values('node', 'z'))

    rtn_value = fields.add_ones('z', at='cell')
    assert_array_equal(rtn_value, np.ones(2))
    assert_is(rtn_value, fields['cell']['z'])
    assert_is(rtn_value, fields.field_values('cell', 'z'))


def test_add_existing_field_default():
    """Test default is to not replace existing field."""
    fields = ModelDataFields()
    fields.new_field_location('node', 12)
    fields.add_empty('z', at='node')

    assert_raises(FieldError, fields.add_empty, 'z', at='node')
    assert_raises(FieldError, fields.add_ones, 'z', at='node')
    assert_raises(FieldError, fields.add_zeros, 'z', at='node')


def test_add_existing_field_with_noclobber():
    """Test noclobber raises an error with an existing field."""
    fields = ModelDataFields()
    fields.new_field_location('node', 12)
    fields.add_empty('z', at='node')

    assert_raises(FieldError, fields.add_empty, 'z', at='node', noclobber=True)
    assert_raises(FieldError, fields.add_ones, 'z', at='node', noclobber=True)
    assert_raises(FieldError, fields.add_zeros, 'z', at='node', noclobber=True)


def test_add_field_with_noclobber():
    """Test noclobber does not raise an error with an new field."""
    fields = ModelDataFields()
    fields.new_field_location('node', 12)

    fields.add_empty('a', at='node', noclobber=True)
    assert_true('a' in fields['node'])

    fields.add_ones('b', at='node', noclobber=True)
    assert_true('b' in fields['node'])

    fields.add_zeros('c', at='node', noclobber=True)
    assert_true('c' in fields['node'])


def test_add_field_with_clobber():
    """Test adding a field with clobber on."""
    fields = ModelDataFields()
    fields.new_field_location('node', 12)

    assert_is_not(fields.add_empty('a', at='node'),
                  fields.add_empty('a', at='node', noclobber=False))
    assert_is_not(fields.add_ones('b', at='node'),
                  fields.add_ones('b', at='node', noclobber=False))
    assert_is_not(fields.add_zeros('c', at='node'),
                  fields.add_zeros('c', at='node', noclobber=False))


def test_getitem():
    fields = ModelDataFields()
    fields.new_field_location('node', 12)

    assert_dict_equal(dict(), fields['node'])
    assert_raises(GroupError, lambda k: fields[k], 'cell')
    assert_raises(KeyError, lambda k: fields[k], 'cell')


def test_at_attribute():
    fields = ModelDataFields()
    fields.new_field_location('node', 12)

    assert_dict_equal(dict(), fields.at_node)
    assert_raises(AttributeError, lambda: fields.at_cell)

    fields.add_ones('z', at='node')
    assert_array_equal(np.ones(12), fields.at_node['z'])


def test_has_group():
    fields = ModelDataFields()
    fields.new_field_location('node', 12)

    assert_true(fields.has_group('node'))
    assert_false(fields.has_group('cell'))


def test_delete_field():
    fields = ModelDataFields()
    fields.new_field_location('link', 17)

    assert_dict_equal(dict(), fields.at_link)
    assert_raises(AttributeError, lambda: fields.at_node)

    fields.add_zeros('link', 'vals')
    assert_array_equal(np.zeros(17), fields.at_link['vals'])

    assert_raises(KeyError, lambda: fields.delete_field('node', 'vals'))
    fields.delete_field('link', 'vals')
    assert_raises(KeyError, lambda: fields.field_units('link', 'vals'))
    assert_raises(KeyError, lambda: fields.at_link['vals'])



#! /usr/bin/env python
"""
Unit tests for landlab.model_parameter_dictionary
"""

from nose.tools import assert_equal, assert_true, assert_raises, with_setup
try:
    from nose.tools import assert_is_instance
except ImportError:
    from landlab.testing.tools import assert_is_instance
import os
import tempfile
import numpy as np

from landlab import ModelParameterDictionary
from landlab.core.model_parameter_dictionary import (MissingKeyError,
                                                     ParameterValueError)


def setup():
    from six import StringIO

    _TEST_PARAM_DICT_FILE = u"""
    # A Comment
    INT_VAL:
    1

    FLOAT_VAL: # A Comment
    2.2
    STRING_VAL:
    The Landlab

    TRUE_BOOL_VAL:
    True
    FALSE_BOOL_VAL:
    False
    """

    param_file = StringIO(_TEST_PARAM_DICT_FILE)
    param_dict = ModelParameterDictionary()
    param_dict.read_from_file(param_file)
    globals().update({
        'param_dict': param_dict,
        '_TEST_PARAM_DICT_FILE': _TEST_PARAM_DICT_FILE
    })


@with_setup(setup)
def test_read_file():
    all_keys = set([
        'FLOAT_VAL', 'INT_VAL', 'STRING_VAL', 'TRUE_BOOL_VAL',
        'FALSE_BOOL_VAL',
    ])
    param_list = set(param_dict.params())

    assert_equal(param_list, all_keys)


@with_setup(setup)
def test_read_file_name():
    (prm_fd, prm_file_name) = tempfile.mkstemp()

    prm_file = os.fdopen(prm_fd, 'w')
    prm_file.write(_TEST_PARAM_DICT_FILE)
    prm_file.close()

    param_dict = ModelParameterDictionary()
    param_dict.read_from_file(prm_file_name)

    os.remove(prm_file_name)

    all_keys = set([
        'FLOAT_VAL', 'INT_VAL', 'STRING_VAL', 'TRUE_BOOL_VAL',
        'FALSE_BOOL_VAL',
    ])
    param_list = set(param_dict.params())

    assert_equal(param_list, all_keys)


@with_setup(setup)
def test_read_file_like_twice():
    from six import StringIO
    param_file = StringIO(_TEST_PARAM_DICT_FILE)
    param_dict_1 = ModelParameterDictionary()
    param_dict_2 = ModelParameterDictionary()

    param_dict_1.read_from_file(param_file)
    param_dict_2.read_from_file(param_file)


@with_setup(setup)
def test_read_int():
    assert_equal(param_dict.read_int('INT_VAL'), 1)

    assert_raises(ParameterValueError, param_dict.read_int, 'FLOAT_VAL')

    assert_raises(MissingKeyError, param_dict.read_int, 'MISSING_INT')

    assert_equal(param_dict.read_int('MISSING_INT', 2), 2)


@with_setup(setup)
def test_get_int():
    assert_equal(param_dict.get('INT_VAL', ptype=int), 1)

    assert_raises(ParameterValueError, param_dict.get, 'FLOAT_VAL', ptype=int)

    assert_raises(MissingKeyError, param_dict.get, 'MISSING_INT', ptype=int)


@with_setup(setup)
def test_set_default():
    param_dict.setdefault('MISSING_INT', 2)
    assert_equal(param_dict.read_int('MISSING_INT'), 2)


@with_setup(setup)
def test_read_float():
    assert_equal(param_dict.read_float('FLOAT_VAL'), 2.2)
    assert_equal(param_dict.read_float('INT_VAL'), 1)

    assert_raises(ParameterValueError, param_dict.read_float, 'STRING_VAL')

    assert_raises(MissingKeyError, param_dict.read_float, 'MISSING_FLOAT')


@with_setup(setup)
def test_read_string():
    assert_equal(param_dict.read_string('STRING_VAL'), 'The Landlab')
    assert_equal(param_dict.read_string('INT_VAL'), '1')
    assert_equal(param_dict.read_string('FLOAT_VAL'), '2.2')

    assert_raises(MissingKeyError, param_dict.read_string, 'MISSING_STRING')


@with_setup(setup)
def test_read_bool():
    assert_equal(param_dict.read_bool('TRUE_BOOL_VAL'), True)
    assert_equal(param_dict.read_bool('FALSE_BOOL_VAL'), False)

    assert_raises(MissingKeyError, param_dict.read_bool, 'MISSING_BOOLEAN')

    assert_raises(ParameterValueError, param_dict.read_bool, 'STRING_VAL')


@with_setup(setup)
def test_dict_keys():
    all_keys = set([
        'FLOAT_VAL', 'INT_VAL', 'STRING_VAL', 'TRUE_BOOL_VAL',
        'FALSE_BOOL_VAL',
    ])
    assert_equal(set(param_dict), all_keys)
    for key in all_keys:
        assert_true(key in param_dict)


@with_setup(setup)
def test_dict_index():
    assert_equal(param_dict['INT_VAL'], '1')


def setup_auto_type():
    from six import StringIO

    _TEST_FILE = u"""
# A Comment
INT_VAL:
1
DBL_VAL:
1.2
STR_VAL:
landlab
BOOL_VAL:
true
INT_ARRAY_VAL:
1,2 ,4 ,7

DBL_ARRAY_VAL:
1.,2. ,4. ,7.
    """
    param_dict = ModelParameterDictionary(auto_type=True,
                                          from_file=StringIO(_TEST_FILE))
    globals().update({
        'param_dict': param_dict,
    })


@with_setup(setup_auto_type)
def test_auto_type():
    assert_equal(param_dict['INT_VAL'], 1)
    assert_equal(param_dict['DBL_VAL'], 1.2)
    assert_equal(param_dict['STR_VAL'], 'landlab')
    assert_equal(param_dict['BOOL_VAL'], True)


@with_setup(setup_auto_type)
def test_int_vector():
    val = param_dict['INT_ARRAY_VAL']

    assert_equal(list(val), [1, 2, 4, 7])
    assert_is_instance(val, np.ndarray)
    assert_equal(val.dtype, np.int)


@with_setup(setup_auto_type)
def test_float_vector():
    val = param_dict['DBL_ARRAY_VAL']

    assert_equal(list(val), [1., 2., 4., 7.])
    assert_is_instance(val, np.ndarray)
    assert_equal(val.dtype, np.float)



import numpy as np
from numpy.testing import assert_array_equal, assert_array_almost_equal
from nose.tools import with_setup

from landlab import RasterModelGrid


def setup_unit_grid():
    globals().update({
        '_RMG': RasterModelGrid(4, 5)
    })


def setup_grid():
    globals().update({
        '_RMG': RasterModelGrid((4, 5), spacing=(3, 4))
    })


@with_setup(setup_unit_grid)
def test_unit_spacing():
    lengths_incl_diags = _RMG._create_length_of_link()
    lengths = _RMG.length_of_link
    assert_array_equal(lengths_incl_diags[:_RMG.number_of_links], np.ones(31))
    assert_array_equal(lengths, np.ones(31))
    assert_array_almost_equal(lengths_incl_diags[_RMG.number_of_links:],
                              np.sqrt(2.)*np.ones(24))


@with_setup(setup_grid)
def test_non_unit_spacing():
    assert_array_equal(_RMG._create_length_of_link()[:_RMG.number_of_links],
                       [4., 4., 4., 4.,
                        3., 3., 3., 3., 3.,
                        4., 4., 4., 4.,
                        3., 3., 3., 3., 3.,
                        4., 4., 4., 4.,
                        3., 3., 3., 3., 3.,
                        4., 4., 4., 4.])

    assert_array_equal(_RMG._create_length_of_link()[_RMG.number_of_links:],
                       5.*np.ones(24))


@with_setup(setup_grid)
def test_link_length():
    assert_array_equal(_RMG.length_of_link,
                       [4., 4., 4., 4.,
                        3., 3., 3., 3., 3.,
                        4., 4., 4., 4.,
                        3., 3., 3., 3., 3.,
                        4., 4., 4., 4.,
                        3., 3., 3., 3., 3.,
                        4., 4., 4., 4.])


@with_setup(setup_grid)
def test_active_link_length():
    assert_array_equal(_RMG.length_of_link[_RMG.active_links],
                       [3., 3., 3.,
                        4., 4., 4., 4.,
                        3., 3., 3.,
                        4., 4., 4., 4.,
                        3., 3., 3.])



import numpy as np
from numpy.testing import assert_array_equal
from nose.tools import with_setup, raises
try:
    from nose.tools import assert_tuple_equal
except ImportError:
    from landlab.testing.tools import assert_tuple_equal


from landlab import RasterModelGrid


def setup_default_grid():
    globals().update({
        'rmg': RasterModelGrid(4, 5)
    })


def setup_lon_lat_grid():
    globals().update({
        'rmg': RasterModelGrid(4, 5,
                               axis_name=['longitude', 'latitude'],
                               axis_units=['degrees_east', 'degrees_north'])
    })


@with_setup(setup_default_grid)
def test_default_names():
    assert_tuple_equal(rmg.axis_name, ('y', 'x'))


@with_setup(setup_lon_lat_grid)
def test_name_keyword():
    assert_tuple_equal(rmg.axis_name, ('longitude', 'latitude'))


@with_setup(setup_lon_lat_grid)
def test_name_setter():
    rmg.axis_name = ('yyy', 'xxx')
    assert_tuple_equal(rmg.axis_name, ('yyy', 'xxx'))


@raises(ValueError)
@with_setup(setup_default_grid)
def test_name_setter_too_few_names():
    rmg.axis_name = ('z', )


@raises(ValueError)
@with_setup(setup_default_grid)
def test_name_setter_too_many_names():
    rmg.axis_name = ('z', 'y', 'x')


@with_setup(setup_default_grid)
def test_default_units():
    assert_tuple_equal(rmg.axis_units, ('-', '-'))


@with_setup(setup_lon_lat_grid)
def test_name_keyword():
    assert_tuple_equal(rmg.axis_units, ('degrees_east', 'degrees_north'))


@with_setup(setup_lon_lat_grid)
def test_name_setter():
    rmg.axis_units = ('mm', 'cm')
    assert_tuple_equal(rmg.axis_units, ('mm', 'cm'))


@raises(ValueError)
@with_setup(setup_default_grid)
def test_name_setter_too_few_units():
    rmg.axis_units = ('m', )


@raises(ValueError)
@with_setup(setup_default_grid)
def test_name_setter_too_many_units():
    rmg.axis_units = ('m', 'cm', 'mm')



"""
The Basic Modeling Interface.
"""


class Error(Exception):
    """Base class for BMI exceptions"""
    pass


class FatalError(Exception):
    """
    Raise this exception if an unrecoverable error was found
    """
    pass


class BadVarNameError(Error):
    """Exception to indicate a bad input/output variable name"""

    def __init__(self, name):
        super(BadVarNameError, self).__init__()
        self.name = name

    def __str__(self):
        return self.name


class MissingModelAttributeError(Error):
    """
    Raise this exception if a component is missing a required attribute.
    """

    def __init__(self, attrib):
        super(MissingModelAttributeError, self).__init__()
        self.attrib = attrib

    def __str__(self):
        return self.attrib


class TimeBoundsError(Error):
    """
    Raise this exception if a component updates beyond its time horizon
    """
    pass


class BmiGridType(int):
    """
    Base type to indicate the type of a BMI model's grid.

    :code: Grid type code as an int
    :name: Name of the grid type as a string
    """
    def __new__(cls, code, name):
        obj = super(BmiGridType, cls).__new__(cls, code)
        obj.name = name
        return obj

    def __str__(self):
        return self.name

    def __repr__(self):
        return 'BmiGridType(%d, %s)' % (self, self.name)


GRID_TYPE_UNKNOWN = BmiGridType(-1, 'Unknown')
GRID_TYPE_NONE = BmiGridType(0, 'No grid')
GRID_TYPE_UNIFORM = BmiGridType(1, 'Uniform rectilinear')
GRID_TYPE_RECTILINEAR = BmiGridType(2, 'Rectilinear')
GRID_TYPE_STRUCTURED = BmiGridType(3, 'Structured')
GRID_TYPE_UNSTRUCTURED = BmiGridType(4, 'Unstructured')


class BmiBase(object):
    """
    Definition of the Basic Modeling Interface
    """

    def initialize(self, file_name):
        """
        Initialize model.

        :file_name: String of configuration file
        """
        pass

    def update(self, **kwds):
        """
        Update model by one time step.
        """
        pass

    def finalize(self):
        """
        Clean-up model
        """
        pass

    def get_input_var_names(self):
        """
        Get names of input variables to the model as standard names.

        :returns: A list of input standard names as strings
        """
        pass

    def get_output_var_names(self):
        """
        Get names of output variables to the model as standard names.

        :returns: A list of output standard names as strings
        """
        pass

    def get_var_type(self, var_name):
        """
        Get type of an exchange item.
        """
        pass

    def get_var_units(self, var_name):
        """
        Get units of an exchange item.
        """
        pass

    def get_var_rank(self, var_name):
        """
        Rank of exchange item.
        """
        pass

    def get_time_step(self):
        """
        Model time step.
        """
        pass

    def get_start_time(self):
        """
        Model start time.
        """
        pass

    def get_current_time(self):
        """
        Current time of model.
        """
        pass

    def get_end_time(self):
        """
        Model stop time.
        """
        pass


class BmiExtendedBase(object):
    """
    An extension interface for a BMI.
    """

    def update_until(self, time):
        """
        Update model until some time.

        :time: Update duration
        """
        pass

    def run_model(self):
        """
        Initialize, run, and finalize a model.
        """
        pass


class BmiUnstructured(object):
    """
    BMI for a model that uses an unstructured grid.
    """

    def get_x(self, name):
        """
        Get x-coordinates of grid nodes.
        """
        pass

    def get_y(self, name):
        """
        Get y-coordinates of grid nodes.
        """
        pass

    def get_connectivity(self, name):
        """
        Get cell connectivity.
        """
        pass

    def get_offset(self, name):
        """
        Get cell offset.
        """
        pass


class BmiStructured(object):
    """
    BMI for a model that uses a structured grid.
    """

    def get_grid_shape(self, name):
        """
        Get shape of grid for variable, name.

        :name: Standard name
        """
        pass

    def get_x(self, name):
        """
        Get x-coordinates of grid nodes.
        """
        pass

    def get_y(self, name):
        """
        Get y-coordinates of grid nodes.
        """
        pass


class BmiRectilinear(object):
    """
    BMI for a model that uses a rectilinear grid.
    """

    def get_grid_shape(self, name):
        """
        Get shape of grid for variable, name.

        :name: Standard name
        """
        pass

    def get_columns(self, name):
        """
        Get coordinates of grid columns.
        """
        pass

    def get_rows(self, name):
        """
        Get coordinates of grid rows.
        """
        pass


class BmiUniformRectilinear(object):
    """
    BMI for a model that exposes a uniform rectilinear grid.
    """

    def get_grid_shape(self, name):
        """
        Get shape of grid for variable, name.

        :name: Standard name
        """
        pass

    def get_grid_spacing(self, name):
        """
        Get spacing of grid for variable, name.

        :name: Standard name
        """
        pass

    def get_grid_origin(self, name):
        """
        Get origin of grid for variable, name.

        :name: Standard name
        """
        pass


class BmiNoGrid(object):
    """
    BMI for a model that does not have a grid.
    """



import numpy as np
from numpy.testing import assert_array_equal
from nose.tools import with_setup, assert_true, assert_false
try:
    from nose.tools import assert_tuple_equal
except ImportError:
    from landlab.testing.tools import assert_tuple_equal

from landlab import RasterModelGrid


def setup_grid():
    globals().update({
        'rmg': RasterModelGrid(4, 5)
    })


def test_boundary_node():
    rmg = RasterModelGrid(5, 6)
    assert_true(rmg.node_has_boundary_neighbor(0))
    assert_false(rmg.node_has_boundary_neighbor(14))


@with_setup(setup_grid)
def test_last_index():
    assert_true(rmg.node_has_boundary_neighbor(-1))


@with_setup(setup_grid)
def test_id_as_list():
    assert_array_equal(rmg.node_has_boundary_neighbor([-1, 0]),
                       np.array([True, True]))


@with_setup(setup_grid)
def test_id_as_array():
    assert_array_equal(rmg.node_has_boundary_neighbor(np.arange(20)),
                       np.array([True, True, True, True, True,
                                 True, True, True, True, True,
                                 True, True, True, True, True,
                                 True, True, True, True, True]))


def test_id_as_array_with_one_interior():
    rmg = RasterModelGrid(5, 5)
    assert_array_equal(rmg.node_has_boundary_neighbor(np.arange(25)),
                       np.array([True, True,  True, True, True,
                                 True, True,  True, True, True,
                                 True, True, False, True, True,
                                 True, True,  True, True, True,
                                 True, True,  True, True, True]))



#! /usr/bin/env python
from .grouped import ModelDataFields, GroupSizeError


_GROUPS = ('node', 'cell', 'link', 'face', 'core_node', 'core_cell',
           'active_link', 'active_face', )


class ModelDataFieldsMixIn(ModelDataFields):

    """Mix-in that provides un-sized fields.

    Inherit from this class to provide fields that do not have to be given
    a size on instantiation. The size of a particular group will only be set
    after the first time a group is accessed (a field is added to a group or
    an array is created for a group). The size of that group will be the
    size as returned by the `number_of_elements` method.

    This mix-in assumes it is being added to a class that implements a method
    that, given an element name as a string, returns the number of elements
    for that group. A `RasterModelGrid` is an excellent example of this.

    >>> from landlab import RasterModelGrid
    >>> grid = RasterModelGrid((4, 5))
    >>> grid.number_of_elements('node')
    20

    Examples
    --------

    A `RasterModelGrid` is an example of a class that inherits from
    `ModelDataFieldsMixIn`.

    >>> from landlab import RasterModelGrid, CLOSED_BOUNDARY
    >>> grid = RasterModelGrid((4, 5))

    After we initially create the grid, it has a number of active links.
    However, since we haven't assigned or created a field based on the number,
    the field is un-sized.

    >>> grid.number_of_elements('active_link')
    17
    >>> grid.at_active_link.size is None
    True

    If we change the status of some of the nodes, the number of active links
    changes. If we now create an array based on the current number of active
    links, that number *is now fixed*.

    >>> grid.status_at_node[:5] = CLOSED_BOUNDARY
    >>> grid.number_of_elements('active_link')
    14
    >>> grid.ones(at='active_link', dtype=int)
    array([1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1])
    >>> grid.at_active_link.size
    14

    Because the number of active links is fixed in our collection of fields,
    if we change the number of active links in our grid, an exception is
    raised if we try to create a new active link array.

    >>> grid.status_at_node[-5:] = CLOSED_BOUNDARY
    >>> grid.number_of_elements('active_link')
    11
    >>> grid.ones(at='active_link', dtype=int)
    ...     # doctest: +IGNORE_EXCEPTION_DETAIL
    Traceback (most recent call last):
    GroupSizeError: number of active_link elements has changed. (was = 14, now=11)
    """

    def __init__(self, **kwds):
        super(ModelDataFieldsMixIn, self).__init__(**kwds)
        for group in _GROUPS:
            ModelDataFields.new_field_location(self, group)

    def new_field_location(self, group, size=None):
        raise AttributeError(
            "'ModelDataFieldsMixIn' object has no attribute "
            "'new_field_location'")

    def empty(self, *args, **kwds):
        """Array, filled with unititialized values, for a given element.

        Returns a numpy array of uninitialized values that is the same length
        as the number of nodes in the grid. Use the *centering* keyword to
        return an array for other elements of the grid. *centering* is a
        string that is one of *node*, *cell*, *link*, or *face*.

        All other keywords are the same as for the numpy zeros function.

        Parameters
        ----------
        centering : str, optional
            Grid element on which the values are defined.

        Returns
        -------
        ndarray
            A newly-allocated array.

        Examples
        --------
        >>> from landlab import RasterModelGrid
        >>> grid = RasterModelGrid((4, 5))
        >>> len(grid.empty())
        20
        """
        if len(args) == 0:
            group = kwds.pop('at', kwds.pop('centering', 'node'))
        else:
            group = args[0]

        n_elements = self.number_of_elements(group)

        if self[group].size is None:
            self[group].size = n_elements
        elif self[group].size != n_elements:
            raise GroupSizeError(group, self[group].size, n_elements)

        return ModelDataFields.empty(self, group, **kwds)

    def ones(self, *args, **kwds):
        """Array, filled with ones, for a given element.

        Returns a numpy array of ones that is the same length as the number
        of nodes in the grid. Use the *centering* keyword to return an
        array for other elements of the grid. *centering* is a string that is
        one of *node*, *cell*, *link*, or *face*.

        All other keywords are the same as for the numpy zeros function.

        Parameters
        ----------
        centering : str, optional
            Grid element on which the values are defined.

        Returns
        -------
        ndarray
            A newly-allocated array.

        Examples
        --------
        >>> from landlab import RasterModelGrid
        >>> grid = RasterModelGrid((4, 5))
        >>> grid.zeros(dtype=int) # doctest: +NORMALIZE_WHITESPACE
        array([0, 0, 0, 0, 0,
               0, 0, 0, 0, 0,
               0, 0, 0, 0, 0,
               0, 0, 0, 0, 0])
        >>> grid.zeros('cell') # doctest: +NORMALIZE_WHITESPACE
        array([ 0., 0., 0.,
                0., 0., 0.])
        """
        if len(args) == 0:
            group = kwds.pop('at', kwds.pop('centering', 'node'))
        else:
            group = args[0]

        n_elements = self.number_of_elements(group)

        if self[group].size is None:
            self[group].size = n_elements
        elif self[group].size != n_elements:
            raise GroupSizeError(group, self[group].size, n_elements)

        return ModelDataFields.ones(self, group, **kwds)

    def zeros(self, *args, **kwds):
        """Array, filled with zeros, for a given element.

        Returns a numpy array of zeros that is the same length as the number
        of nodes in the grid. Use the *centering* keyword to return an
        array for other elements of the grid. *centering* is a string that is
        one of *node*, *cell*, *link*, or *face*.

        All other keywords are the same as for the numpy zeros function.

        Parameters
        ----------
        centering : str, optional
            Grid element on which the values are defined.

        Returns
        -------
        ndarray
            A newly-allocated array.

        Examples
        --------
        >>> from landlab import RasterModelGrid
        >>> grid = RasterModelGrid((4, 5))
        >>> grid.zeros()
        array([ 0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,
                0.,  0.,  0.,  0.,  0.,  0.,  0.])
        """
        if len(args) == 0:
            group = kwds.pop('at', kwds.pop('centering', 'node'))
        else:
            group = args[0]

        n_elements = self.number_of_elements(group)

        if self[group].size is None:
            self[group].size = n_elements
        elif self[group].size != n_elements:
            raise GroupSizeError(group, self[group].size, n_elements)

        return ModelDataFields.zeros(self, group, **kwds)



"""Test the flow router component.

This just tests the functionality of the router in toto - no attempt is made
to test the submodules.
Sinks are tested as part of the lake_mapper.

@author: dejh
"""
# Created on Thurs Nov 12, 2015
import os

from six.moves import range

import numpy as np
from numpy.testing import assert_array_equal, assert_array_almost_equal

from nose.tools import (with_setup, assert_true, assert_false, assert_raises,
                        assert_almost_equal, assert_equal)
try:
    from nose.tools import (assert_is, assert_set_equal, assert_dict_equal)
except ImportError:
    from landlab.testing.tools import (assert_is, assert_set_equal,
                                       assert_dict_equal)

import landlab
from landlab import RasterModelGrid, RadialModelGrid, FieldError
from landlab.components.flow_routing import FlowRouter
from landlab import CLOSED_BOUNDARY
from landlab import BAD_INDEX_VALUE as XX


_THIS_DIR = os.path.abspath(os.path.dirname(__file__))


def setup_dans_grid1():
    """
    Create a 5x5 test grid.
    This is a sheet flow test.
    """
    global fr, mg, infile
    global z, Q_in, A_target, frcvr_target, upids_target, Q_target, \
        steepest_target, links2rcvr_target

    mg = RasterModelGrid((5, 5), spacing=(10., 10.))

    infile = os.path.join(_THIS_DIR, 'test_fr_input.txt')

    z = mg.node_x.copy()

    Q_in = np.full(25, 2.)

    A_target = np.array([0.,  0.,  0.,  0.,  0.,
                         3.,  3.,  2.,  1.,  0.,
                         3.,  3.,  2.,  1.,  0.,
                         3.,  3.,  2.,  1.,  0.,
                         0.,  0.,  0.,  0.,  0.])*100.

    frcvr_target = np.array([0,  1,  2,  3,  4,
                             5,  5,  6,  7,  9,
                            10, 10, 11, 12, 14,
                            15, 15, 16, 17, 19,
                            20, 21, 22, 23, 24])

    upids_target = np.array([0,  1,  2,  3,  4,
                             5,  6,  7,  8,  9,
                            10, 11, 12, 13, 14,
                            15, 16, 17, 18, 19,
                            20, 21, 22, 23, 24])

    links2rcvr_target = np.full(25, XX)
    links2rcvr_target[mg.core_nodes] = np.array([ 9, 10, 11,
                                                 18, 19, 20,
                                                 27, 28, 29])

    Q_target = A_target * 2.  # only once Q_in is used

    steepest_target = np.array([0.,  0.,  0.,  0.,  0.,
                                0.,  1.,  1.,  1.,  0.,
                                0.,  1.,  1.,  1.,  0.,
                                0.,  1.,  1.,  1.,  0.,
                                0.,  0.,  0.,  0.,  0.])

    mg.add_field('node', 'topographic__elevation', z, units='-')


def setup_dans_grid2():
    """
    Create a 5x5 test grid.
    This tests more complex routing, with diffs between D4 & D8.
    """
    global fr, mg, infile
    global z, A_target_D8, A_target_D4, frcvr_target_D8, frcvr_target_D4, \
        upids_target_D8, upids_target_D4, steepest_target_D8, \
        steepest_target_D4, links2rcvr_target_D8, links2rcvr_target_D4

    mg = RasterModelGrid((5, 5), spacing=(10., 10.))

    infile = os.path.join(_THIS_DIR, 'test_fr_input.txt')

    z = np.array([7.,  7.,  7.,  7.,  7.,
                  7.,  5., 3.2,  6.,  7.,
                  7.,  2.,  3.,  5.,  7.,
                  7.,  1., 1.9,  4.,  7.,
                  7.,  0.,  7.,  7.,  7.])

    A_target_D8 = np.array([0.,     0.,     0.,     0.,     0.,
                            0.,   100.,   200.,   100.,     0.,
                            0.,   400.,   100.,   100.,     0.,
                            0.,   600.,   300.,   100.,     0.,
                            0.,   900.,     0.,     0.,     0.])

    A_target_D4 = np.array([0.,     0.,     0.,     0.,     0.,
                            0.,   100.,   200.,   100.,     0.,
                            0.,   200.,   400.,   100.,     0.,
                            0.,   900.,   600.,   100.,     0.,
                            0.,   900.,     0.,     0.,     0.])

    frcvr_target_D8 = np.array([0, 1, 2, 3, 4, 5, 11, 11, 7, 9, 10, 16, 16, 17,
                                14, 15, 21, 21, 17, 19, 20, 21, 22, 23, 24])

    frcvr_target_D4 = np.array([ 0,  1,  2,  3,  4, 
                                 5, 11, 12,  7,  9,
                                10, 16, 17, 12, 14,
                                15, 21, 16, 17, 19,
                                20, 21, 22, 23, 24])

    upids_target_D8 = np.array([0, 1, 2, 3, 4, 5, 9, 10, 14, 15, 19, 20, 21,
                                16, 11, 6, 7, 8, 12, 17, 13, 18, 22, 23, 24])

    upids_target_D4 = np.array([0, 1, 2, 3, 4, 5, 9, 10, 14, 15, 19, 20, 21,
                                16, 11, 6, 17, 12, 7, 8, 13, 18, 22, 23, 24])

    links2rcvr_target_D8 = np.full(25, XX)
    links2rcvr_target_D8[mg.core_nodes] = np.array([14, 51, 11,
                                                    23, 59, 61,
                                                    32, 67, 29])

    links2rcvr_target_D4 = np.full(25, XX)
    links2rcvr_target_D4[mg.core_nodes] = np.array([14, 15, 11,
                                                    23, 24, 20,
                                                    32, 28, 29])

    steepest_target_D8 = np.array([0., 0., 0., 0., 0.,
                                   0., 0.3, 0.08485281, 0.28, 0.,
                                   0., 0.1, 0.14142136, 0.21920310, 0.,
                                   0., 0.1, 0.13435029, 0.21,  0.,
                                   0., 0., 0., 0.,  0.])

    steepest_target_D4 = np.array([0., 0., 0., 0., 0.,
                                   0., 0.3, 0.02, 0.28, 0.,
                                   0., 0.1, 0.11, 0.2, 0.,
                                   0., 0.1, 0.09, 0.21, 0.,
                                   0., 0., 0., 0., 0.])

    mg.add_field('node', 'topographic__elevation', z, units='-')


def setup_internal_closed():
    """
    Create a 6x5 test grid, but with two internal nodes closed.
    This is a sheet flow test.
    """
    global fr, mg
    global z, Q_in, A_target, frcvr_target, upids_target, Q_target, \
        steepest_target, links2rcvr_target

    mg = RasterModelGrid((6, 5), spacing=(10., 10.))

    mg.set_closed_boundaries_at_grid_edges(True, True, False, True)
    mg.status_at_node[7] = CLOSED_BOUNDARY
    mg.status_at_node[16] = CLOSED_BOUNDARY

    z = mg.node_x.copy()

    Q_in = np.full(25, 2.)

    A_target = np.array([0.,  0.,  0.,  0.,  0.,
                         1.,  1.,  0.,  1.,  0.,
                         6.,  6.,  3.,  1.,  0.,
                         0.,  0.,  2.,  1.,  0.,
                         3.,  3.,  2.,  1.,  0.,
                         0.,  0.,  0.,  0.,  0.])*100.

    frcvr_target = np.array([0,  1,  2,  3,  4,
                             5,  5,  7, 12,  9,
                            10, 10, 11, 12, 14,
                            15, 16, 11, 17, 19,
                            20, 20, 21, 22, 24,
                            25, 26, 27, 28, 29])

    links2rcvr_target = np.full(mg.number_of_nodes, XX)
    links2rcvr_target[mg.core_nodes] = np.array([ 9, 62,
                                                 18, 19, 20,
                                                 67, 29,
                                                 36, 37, 38])

    steepest_target = np.array([0.,  0.,  0.,  0.,  0.,
                                0.,  1.,  0.,  0.,  0.,
                                0.,  1.,  1.,  1.,  0.,
                                0.,  0.,  0.,  1.,  0.,
                                0.,  1.,  1.,  1.,  0.,
                                0.,  0.,  0.,  0.,  0.])
    steepest_target[np.array([8, 17])] = 1./np.sqrt(2.)

    mg.add_field('node', 'topographic__elevation', z, units='-')


def setup_voronoi():
    """
    Setup a simple 20 point Voronoi Delaunay grid (radial for ease)
    """
    global fr, vmg
    global A_target_core, A_target_outlet
    vmg = RadialModelGrid(2, dr=2.)
    z = np.full(20, 10., dtype=float)
    #vmg.status_at_node[8:] = CLOSED_BOUNDARY
    all_bounds_but_one = np.array((0, 1, 2, 3, 4, 7, 11, 15, 16, 17, 18, 19))
    vmg.status_at_node[all_bounds_but_one] = CLOSED_BOUNDARY
    #z[7] = 0.  # outlet
    z[12] = 0.  # outlet
    #inner_elevs = (3., 1., 4., 5., 6., 7., 8.)
    inner_elevs = (8., 7., 3., 1., 6., 4., 5.)
    #z[:7] = np.array(inner_elevs)
    z[vmg.core_nodes] = np.array(inner_elevs)
    vmg.add_field('node', 'topographic__elevation', z, units='-')
    fr = FlowRouter(vmg)

#    nodes_contributing = [np.array([0, 3, 4, 5]),
#                          np.array([0, 1, 2, 3, 4, 5, 6]),
#                          np.array([2, ]),
#                          np.array([3, ]),
#                          np.array([4, ]),
#                          np.array([5, ]),
#                          np.array([6, ])]

    # The follow list contains arrays with the IDs of cells contributing flow
    # to nodes 5, 6, 8, 9, 10, 13, and 14, respectively (which correspond to
    # cells 0-6)
    cells_contributing = [np.array([0, ]),
                          np.array([1, ]),
                          np.array([1, 2, 4, 6]),
                          np.array([0, 1, 2, 3, 4, 5, 6]),
                          np.array([4, ]),
                          np.array([5, ]),
                          np.array([6, ])]

    A_target_core = np.zeros(vmg.number_of_core_nodes)
    for i in range(7):
        A_target_core[i] = vmg.area_of_cell[cells_contributing[i]].sum()
    A_target_outlet = vmg.area_of_cell.sum()


def setup_voronoi_closedinternal():
    """
    Close one of the formerly core internal nodes.
    """
    global fr, vmg
    global A_target_internal, A_target_outlet
    vmg = RadialModelGrid(2, dr=2.)
    z = np.full(20, 10., dtype=float)
#    vmg.status_at_node[8:] = CLOSED_BOUNDARY
#    vmg.status_at_node[0] = CLOSED_BOUNDARY  # new internal closed
#    z[7] = 0.  # outlet
#    inner_elevs = (3., 1., 4., 5., 6., 7., 8.)
#    z[:7] = np.array(inner_elevs)
    all_bounds_but_one = np.array((0, 1, 2, 3, 4, 7, 11, 15, 16, 17, 18, 19))
    vmg.status_at_node[all_bounds_but_one] = CLOSED_BOUNDARY
    vmg.status_at_node[8] = CLOSED_BOUNDARY  # new internal closed
    z[12] = 0.  # outlet
    inner_elevs = (8., 7., 1., 6., 4., 5.)
    z[vmg.core_nodes] = np.array(inner_elevs)
    vmg.add_field('node', 'topographic__elevation', z, units='-')
    fr = FlowRouter(vmg)

#    nodes_contributing = [[],
#                          np.array([1, 2, 3, 4, 5, 6]),
#                          np.array([2, 3, 4, 5]),
#                          np.array([3, 4, 5]),
#                          np.array([4, 5]),
#                          np.array([5, ]),
#                          np.array([6, ])]

    cells_contributing = [np.array([0, ]),
                          np.array([1, ]),
                          np.array([0, 1, 3, 4, 5, 6]),
                          np.array([1, 4]),
                          np.array([1, 4, 5, 6]),
                          np.array([1, 4, 6])]

    A_target_internal = np.zeros(vmg.number_of_core_nodes, dtype=float)
    for i in range(6):
        A_target_internal[i] = vmg.area_of_cell[cells_contributing[i]].sum()
#        A_target_internal[i] = vmg.area_of_cell[nodes_contributing[i]].sum()
    A_target_outlet = vmg.area_of_cell[vmg.cell_at_node[vmg.core_nodes]].sum()


@with_setup(setup_dans_grid1)
def test_check_fields():
    """Check to make sure the right fields have been created."""
    fr = FlowRouter(mg)
    assert_array_equal(z, mg.at_node['topographic__elevation'])
    assert_array_equal(np.zeros(25), mg.at_node['drainage_area'])
    assert_array_equal(np.ones(25), mg.at_node['water__unit_flux_in'])

    fr = FlowRouter(mg, infile)
    assert_array_equal(np.full(25, 2.), mg.at_node['water__unit_flux_in'])


@with_setup(setup_dans_grid1)
def test_check_field_input():
    """Check we can successfully pass water__discharge_in."""
    mg.add_field('node', 'water__unit_flux_in',
                 np.full(25, 3.), units='m**3/s')
    fr = FlowRouter(mg)
    assert_array_equal(np.full(25, 3.), mg.at_node['water__unit_flux_in'])
    fr = FlowRouter(mg, infile)
    assert_array_equal(np.full(25, 2.), mg.at_node['water__unit_flux_in'])


@with_setup(setup_dans_grid1)
def test_accumulate_D8():
    """Test accumulation works for D8 in a simple scenario."""
    fr = FlowRouter(mg)
    fr.route_flow()
    assert_array_equal(A_target, mg.at_node['drainage_area'])
    assert_array_equal(frcvr_target, mg.at_node['flow__receiver_node'])
    assert_array_equal(upids_target, mg.at_node['flow__upstream_node_order'])
    assert_array_equal(links2rcvr_target, mg.at_node['flow__link_to_receiver_node'])
    assert_array_equal(A_target, mg.at_node['water__discharge'])
    assert_array_equal(steepest_target,
                       mg.at_node['topographic__steepest_slope'])


@with_setup(setup_dans_grid1)
def test_variable_Qin():
    """Test variable Qin field."""
    Qin_local = np.zeros(25, dtype=float)
    Qin_local[13] = 2.
    mg.add_field('node', 'water__unit_flux_in',
                 Qin_local, units='m**3/s')
    fr = FlowRouter(mg)
    fr.route_flow()
    Qout_local = np.zeros_like(Qin_local)
    Qout_local[10:14] = 200.
    assert_array_equal(Qout_local, mg.at_node['water__discharge'])
    assert_array_equal(A_target, mg.at_node['drainage_area'])
    # note that A DOES NOT CHANGE when messing with Q_in


@with_setup(setup_dans_grid2)
def test_irreg_topo():
    """Test D8 routing on a toy irregular topo."""
    fr = FlowRouter(mg)
    fr.route_flow()
    assert_array_equal(A_target_D8, mg.at_node['drainage_area'])
    assert_array_equal(frcvr_target_D8, mg.at_node['flow__receiver_node'])
    assert_array_equal(upids_target_D8, mg.at_node['flow__upstream_node_order'])
    assert_array_equal(links2rcvr_target_D8,
                       mg.at_node['flow__link_to_receiver_node'])
    assert_array_almost_equal(steepest_target_D8,
                              mg.at_node['topographic__steepest_slope'])


@with_setup(setup_dans_grid2)
def test_irreg_topo_old():
    """
    Test D4 routing on a toy irregular topo, old style, where 'method' is
    passed to the run method, not the init.
    """
    fr = FlowRouter(mg)
    fr.route_flow(method='D4')
    assert_array_equal(A_target_D4, mg.at_node['drainage_area'])
    assert_array_equal(frcvr_target_D4, mg.at_node['flow__receiver_node'])
    assert_array_equal(upids_target_D4, mg.at_node['flow__upstream_node_order'])
    assert_array_equal(links2rcvr_target_D4,
                       mg.at_node['flow__link_to_receiver_node'])
    assert_array_almost_equal(steepest_target_D4,
                              mg.at_node['topographic__steepest_slope'])


@with_setup(setup_dans_grid2)
def test_irreg_topo_new():
    """Test D4 routing on a toy irregular topo. 'method' passed to init."""
    fr = FlowRouter(mg, method='D4')
    fr.route_flow()
    assert_array_equal(A_target_D4, mg.at_node['drainage_area'])
    assert_array_equal(frcvr_target_D4, mg.at_node['flow__receiver_node'])
    assert_array_equal(upids_target_D4, mg.at_node['flow__upstream_node_order'])
    assert_array_equal(links2rcvr_target_D4,
                       mg.at_node['flow__link_to_receiver_node'])
    assert_array_almost_equal(steepest_target_D4,
                              mg.at_node['topographic__steepest_slope'])


@with_setup(setup_internal_closed)
def test_internal_closed():
    """Test closed nodes in the core of the grid."""
    fr = FlowRouter(mg)
    fr.route_flow()
    assert_array_almost_equal(A_target, mg.at_node['drainage_area'])
    assert_array_equal(frcvr_target, mg.at_node['flow__receiver_node'])
    assert_array_equal(links2rcvr_target, mg.at_node['flow__link_to_receiver_node'])
    assert_array_almost_equal(A_target, mg.at_node['water__discharge'])
    assert_array_almost_equal(steepest_target,
                              mg.at_node['topographic__steepest_slope'])


@with_setup(setup_voronoi)
def test_voronoi():
    """Test routing on a (radial) voronoi."""
    fr.route_flow()
    assert_array_almost_equal(vmg.at_node['drainage_area'][vmg.core_nodes],
                              A_target_core)
    assert_almost_equal(vmg.at_node['drainage_area'][12], A_target_outlet)


@with_setup(setup_voronoi_closedinternal)
def test_voronoi_closedinternal():
    """Test routing on a (radial) voronoi, but with a closed interior node."""
    fr.route_flow()
    # for i in range(vmg.number_of_nodes):
    #     print i, vmg.node_x[i], vmg.node_y[i], vmg.status_at_node[i], \
    #             vmg.at_node['drainage_area'][i], vmg.at_node['flow__receiver_node'][i], \
    #             vmg.at_node['topographic__elevation'][i]

    assert_array_almost_equal(vmg.at_node['drainage_area'][vmg.core_nodes],
                              A_target_internal)
    assert_almost_equal(vmg.at_node['drainage_area'][12], A_target_outlet)



"""Test the SedDepEroder component.

Test the sed dep eroder by turning it over a few times. No attempt has been
made to ensure the solution is stable. Takes a topo already output and runs it
a few more times, to ensure repeatability.
"""
import os

from six.moves import range

import numpy as np
from numpy.testing import assert_array_almost_equal

from landlab import RasterModelGrid
from landlab.components.flow_routing import FlowRouter
from landlab.components.stream_power import SedDepEroder
from landlab import ModelParameterDictionary


_THIS_DIR = os.path.abspath(os.path.dirname(__file__))


def test_sed_dep():
    input_file = os.path.join(_THIS_DIR, 'sed_dep_params.txt')
    inputs = ModelParameterDictionary(input_file, auto_type=True)
    nrows = inputs.read_int('nrows')
    ncols = inputs.read_int('ncols')
    dx = inputs.read_float('dx')
    leftmost_elev = inputs.read_float('leftmost_elevation')
    initial_slope = inputs.read_float('initial_slope')
    uplift_rate = inputs.read_float('uplift_rate')

    runtime = inputs.read_float('total_time')
    dt = inputs.read_float('dt')

    nt = int(runtime // dt)
    uplift_per_step = uplift_rate * dt

    mg = RasterModelGrid((nrows, ncols), (dx, dx))

    mg.add_zeros('topographic__elevation', at='node')
    z = np.loadtxt(os.path.join(_THIS_DIR, 'seddepinit.txt'))
    mg['node']['topographic__elevation'] = z

    mg.set_closed_boundaries_at_grid_edges(True, False, True, False)

    fr = FlowRouter(mg)
    sde = SedDepEroder(mg, **inputs)

    for i in range(nt):
        mg.at_node['topographic__elevation'][mg.core_nodes] += uplift_per_step
        mg = fr.route_flow()
        mg, _ = sde.erode(dt)

    z_tg = np.loadtxt(os.path.join(_THIS_DIR, 'seddepz_tg.txt'))

    assert_array_almost_equal(mg.at_node['topographic__elevation'][
        mg.core_nodes], z_tg[mg.core_nodes])

def test_sed_dep_new():
    mg = RasterModelGrid((10, 10), 100.)


#! /usr/bin/env python
"""Unit tests for landlab.io.netcdf module."""

import numpy as np
from nose.tools import assert_equal, assert_true, assert_raises
from nose import SkipTest
from numpy.testing import assert_array_equal

from landlab import RasterModelGrid
from landlab.io.netcdf import write_netcdf, NotRasterGridError, WITH_NETCDF4
from landlab.io.netcdf.read import _get_raster_spacing
from landlab.testing.tools import cdtemp

try:
    import netCDF4 as nc
except ImportError:
    pass


def test_netcdf_write_int64_field_netcdf4():
    """Test write_netcdf with a grid that has an int64 field."""
    field = RasterModelGrid(4, 3)
    field.add_field('node', 'topographic__elevation',
                    np.arange(12, dtype=np.int64))

    with cdtemp() as _:
        write_netcdf('test.nc', field, format='NETCDF4')

        root = nc.Dataset('test.nc', 'r', format='NETCDF4')

        for name in ['topographic__elevation']:
            assert_true(name in root.variables)
            assert_array_equal(root.variables[name][:].flat,
                               field.at_node[name])
            assert_equal(root.variables[name][:].dtype, 'int64')

        root.close()


def test_netcdf_write_uint8_field_netcdf4():
    """Test write_netcdf with a grid that has an uint8 field."""
    field = RasterModelGrid(4, 3)
    field.add_field('node', 'topographic__elevation',
                    np.arange(12, dtype=np.uint8))

    with cdtemp() as _:
        write_netcdf('test.nc', field, format='NETCDF4')

        root = nc.Dataset('test.nc', 'r', format='NETCDF4')

        for name in ['topographic__elevation']:
            assert_true(name in root.variables)
            assert_array_equal(root.variables[name][:].flat,
                               field.at_node[name])
            assert_equal(root.variables[name][:].dtype, 'uint8')

        root.close()


def test_netcdf_write_as_netcdf3_64bit():
    """Test write_netcdf with output format 64-bit netcdf3."""
    from scipy.io import netcdf

    field = RasterModelGrid(4, 3)
    field.add_field('node', 'topographic__elevation', np.arange(12.))
    field.add_field('node', 'uplift_rate', 2. * np.arange(12.))

    with cdtemp() as _:
        write_netcdf('test.nc', field, format='NETCDF3_64BIT')

        f = netcdf.netcdf_file('test.nc', 'r')

        for name in ['topographic__elevation', 'uplift_rate']:
            assert_true(name in f.variables)
            assert_array_equal(f.variables[name][:].flat, field.at_node[name])

        f.close()


def test_netcdf_write_as_netcdf3_classic():
    """Test write_netcdf with output format classic netcdf3."""
    from scipy.io import netcdf

    field = RasterModelGrid(4, 3)
    field.add_field('node', 'topographic__elevation', np.arange(12.))
    field.add_field('node', 'uplift_rate', 2. * np.arange(12.))

    with cdtemp() as _:
        write_netcdf('test.nc', field, format='NETCDF3_CLASSIC')

        f = netcdf.netcdf_file('test.nc', 'r')

        for name in ['topographic__elevation', 'uplift_rate']:
            assert_true(name in f.variables)
            assert_array_equal(f.variables[name][:].flat, field.at_node[name])

        f.close()


def test_netcdf_write():
    """Test generic write_netcdf."""
    if not WITH_NETCDF4:
        raise SkipTest('netCDF4 package not installed')

    field = RasterModelGrid(4, 3)
    field.add_field('node', 'topographic__elevation', np.arange(12.))

    with cdtemp() as _:
        write_netcdf('test.nc', field, format='NETCDF4')
        root = nc.Dataset('test.nc', 'r', format='NETCDF4')

        assert_equal(set(root.dimensions), set(['ni', 'nj', 'nt']))
        assert_equal(len(root.dimensions['ni']), 3)
        assert_equal(len(root.dimensions['nj']), 4)
        assert_true(len(root.dimensions['nt']), 1)
        assert_true(root.dimensions['nt'].isunlimited())

        assert_equal(set(root.variables),
                     set(['x', 'y', 'topographic__elevation']))

        assert_array_equal(root.variables['x'][:].flat,
                           np.array([0., 1., 2., 0., 1., 2., 0., 1., 2.,
                                     0., 1., 2., ]))
        assert_array_equal(root.variables['y'][:].flat,
                           np.array([0., 0., 0., 1., 1., 1., 2., 2., 2.,
                                     3., 3., 3., ]))
        assert_array_equal(root.variables['topographic__elevation'][:].flat,
                           field.at_node['topographic__elevation'])

        root.close()


def test_netcdf_write_as_netcdf4_classic():
    """Test write_netcdf to netcdf4 classic format."""
    if not WITH_NETCDF4:
        raise SkipTest('netCDF4 package not installed')

    field = RasterModelGrid(4, 3)
    field.add_field('node', 'topographic__elevation', np.arange(12.))
    field.add_field('node', 'uplift_rate', np.arange(12.))

    with cdtemp() as _:
        write_netcdf('test.nc', field, format='NETCDF4_CLASSIC')
        root = nc.Dataset('test.nc', 'r', format='NETCDF4_CLASSIC')

        for name in ['topographic__elevation', 'uplift_rate']:
            assert_true(name in root.variables)
            assert_array_equal(root.variables[name][:].flat,
                               field.at_node[name])

        root.close()


def test_netcdf_write_names_keyword_as_list():
    """Test write_netcdf using a list for the *names* keyword."""
    if not WITH_NETCDF4:
        raise SkipTest('netCDF4 package not installed')

    field = RasterModelGrid(4, 3)
    field.add_field('node', 'topographic__elevation', np.arange(12.))
    field.add_field('node', 'uplift_rate', np.arange(12.))

    with cdtemp() as _:
        write_netcdf('test.nc', field, names=['topographic__elevation'],
                     format='NETCDF4')
        root = nc.Dataset('test.nc', 'r', format='NETCDF4')

        assert_true('topographic__elevation' in root.variables)
        assert_true('uplift_rate' not in root.variables)
        assert_array_equal(root.variables['topographic__elevation'][:].flat,
                           field.at_node['topographic__elevation'])

        root.close()


def test_netcdf_write_names_keyword_as_str():
    """Test write_netcdf using a ``str`` for the *names* keyword."""
    if not WITH_NETCDF4:
        raise SkipTest('netCDF4 package not installed')

    field = RasterModelGrid(4, 3)
    field.add_field('node', 'topographic__elevation', np.arange(12.))
    field.add_field('node', 'uplift_rate', np.arange(12.))

    with cdtemp() as _:
        write_netcdf('test.nc', field, names='uplift_rate', format='NETCDF4')
        root = nc.Dataset('test.nc', 'r', format='NETCDF4')

        assert_true('topographic__elevation' not in root.variables)
        assert_true('uplift_rate' in root.variables)
        assert_array_equal(root.variables['uplift_rate'][:].flat,
                           field.at_node['uplift_rate'])

        root.close()


def test_netcdf_write_names_keyword_as_none():
    """Test write_netcdf using ``None`` for the *names* keyword."""
    if not WITH_NETCDF4:
        raise SkipTest('netCDF4 package not installed')

    field = RasterModelGrid(4, 3)
    field.add_field('node', 'topographic__elevation', np.arange(12.))
    field.add_field('node', 'uplift_rate', np.arange(12.))

    with cdtemp() as _:
        write_netcdf('test.nc', field, names=None, format='NETCDF4')
        root = nc.Dataset('test.nc', 'r', format='NETCDF4')

        for name in ['topographic__elevation', 'uplift_rate']:
            assert_true(name in root.variables)
            assert_array_equal(root.variables[name][:].flat,
                               field.at_node[name])

        root.close()


def test_2d_unit_spacing():
    """Test write_netcdf with a 2D grid with unit spacing."""
    (x, y) = np.meshgrid(np.arange(5.), np.arange(4.))

    spacing = _get_raster_spacing((y, x))
    assert_equal(spacing, 1.)


def test_2d_non_unit_spacing():
    """Test _get_raster_spacing with a 2D grid with non-unit spacing."""
    (x, y) = np.meshgrid(np.arange(5.) * 2, np.arange(4.) * 2)

    spacing = _get_raster_spacing((y, x))
    assert_equal(spacing, 2.)


def test_2d_uneven_spacing_axis_0():
    """Test _get_raster_spacing with a 2D grid with uneven spacing in y."""
    (x, y) = np.meshgrid(np.logspace(0., 2., num=5), np.arange(4.))

    assert_raises(NotRasterGridError, _get_raster_spacing, (y, x))


def test_2d_uneven_spacing_axis_1():
    """Test _get_raster_spacing with a 2D grid with uneven spacing in x."""
    (x, y) = np.meshgrid(np.arange(4.), np.logspace(0., 2., num=5))

    assert_raises(NotRasterGridError, _get_raster_spacing, (y, x))


def test_2d_switched_coords():
    """Test _get_raster_spacing with a 2D grid when the spacing is switched."""
    (x, y) = np.meshgrid(np.arange(5.), np.arange(4.))

    spacing = _get_raster_spacing((x, y))
    assert_equal(spacing, 0.)


def test_1d_unit_spacing():
    """Test _get_raster_spacing with a 1D grid with unit spacing."""
    spacing = _get_raster_spacing((np.arange(5.), ))
    assert_equal(spacing, 1.)


def test_1d_non_unit_spacing():
    """Test _get_raster_spacing with a 1D grid with non-unit spacing."""
    spacing = _get_raster_spacing((np.arange(5.) * 2, ))
    assert_equal(spacing, 2.)


def test_1d_uneven_spacing():
    """Test _get_raster_spacing with a 1D grid with uneven spacing in y."""
    assert_raises(NotRasterGridError, _get_raster_spacing,
                  (np.logspace(0., 2., num=5), ))


def test_netcdf_write_at_cells():
    """Test write_netcdf using with cell fields"""
    if not WITH_NETCDF4:
        raise SkipTest('netCDF4 package not installed')

    field = RasterModelGrid((4, 3))
    field.add_field('cell', 'topographic__elevation',
                    np.arange(field.number_of_cells))
    field.add_field('cell', 'uplift_rate', np.arange(field.number_of_cells))

    with cdtemp() as _:
        write_netcdf('test-cells.nc', field, format='NETCDF4')
        root = nc.Dataset('test-cells.nc', 'r', format='NETCDF4')

        for name in ['topographic__elevation', 'uplift_rate']:
            assert_true(name in root.variables)
            assert_array_equal(root.variables[name][:].flat,
                               field.at_cell[name])

        assert_equal(set(root.dimensions), set(['nv', 'ni', 'nj', 'nt']))
        assert_equal(len(root.dimensions['nv']), 4)
        assert_equal(len(root.dimensions['ni']), 1)
        assert_equal(len(root.dimensions['nj']), 2)
        assert_true(len(root.dimensions['nt']), 1)
        assert_true(root.dimensions['nt'].isunlimited())

        assert_equal(set(root.variables),
                     set(['x_bnds', 'y_bnds', 'topographic__elevation',
                          'uplift_rate']))
        root.close()



import sys
import warnings

from .bmi_bridge import wrap_as_bmi
from ..components import COMPONENTS


__all__ = []
for cls in COMPONENTS:
    try:
        as_bmi = wrap_as_bmi(cls)
    except TypeError:
        warnings.warn('unable to wrap class {name}'.format(name=cls.__name__))
    else:
        setattr(sys.modules[__name__], as_bmi.__name__, as_bmi)
        __all__.append(as_bmi.__name__)



# -*- coding: utf-8 -*-
"""
hex_lattice_tectonicizer.py.

Models discrete normal-fault offset on a 2D hex lattice with a rectangular
shape and with one orientation of the nodes being vertical.

The intention here is to use a particle (LCA) model to represent the evolution
of a 2D hillslope, with the hex_lattice_tectonicizer serving to shift the nodes
either upward (simple vertical uplift relative to baselevel), or up and
sideways (representing motion on a fault plane).

Created on Mon Nov 17 08:01:49 2014

@author: gtucker
"""

from landlab import HexModelGrid
from numpy import amax, zeros, arange, array, sqrt

_DEFAULT_NUM_ROWS = 5
_DEFAULT_NUM_COLS = 5
_TAN60 = 1.732


class HexLatticeTectonicizer(object):
    """Handles tectonics and baselevel for CellLab-CTS models.

    This is the base class from which classes to represent particular
    baselevel/fault geometries are derived.

    Examples
    --------
    >>> hlt = HexLatticeTectonicizer()
    >>> hlt.grid.number_of_nodes
    25
    >>> hlt.nr
    5
    >>> hlt.nc
    5
    """

    def __init__(self, grid=None, node_state=None, propid=None, prop_data=None,
                 prop_reset_value=None):
        """
        Create and initialize a HexLatticeTectonicizer.

        Examples
        --------
        >>> from landlab import HexModelGrid
        >>> hg = HexModelGrid(6, 6, shape='rect')
        >>> hlt = HexLatticeTectonicizer()
        >>> hlt.grid.number_of_nodes
        25
        >>> hlt.nr
        5
        >>> hlt.nc
        5
        """
        # If needed, create grid
        if grid is None:
            num_rows = _DEFAULT_NUM_ROWS
            num_cols = _DEFAULT_NUM_COLS
            self.grid = HexModelGrid(num_rows, num_cols, dx=1.0,
                                     orientation='vertical',
                                     shape='rect', reorient_links=True)
        else:
            # Make sure caller passed the right type of grid
            assert (grid.orientation == 'vertical'), \
                   'Grid must have vertical orientation'

            # Keep a reference to the grid
            self.grid = grid

        # If needed, create node-state grid
        if node_state is None:
            self.node_state = self.grid.add_zeros('node', 'node_state_map',
                                                  dtype=int)
        else:
            self.node_state = node_state

        # Remember the # of rows and cols
        self.nr = self.grid.number_of_node_rows
        self.nc = self.grid.number_of_node_columns

        # propid should be either a reference to a CA model's "property id"
        # array, or None
        self.propid = propid
        self.prop_data = prop_data
        self.prop_reset_value = prop_reset_value


class LatticeNormalFault(HexLatticeTectonicizer):
    """Handles normal-fault displacement in CellLab-CTS models.

    Represents a 60 degree, left-dipping normal fault, and handles discrete
    offsets for a hex grid that has vertical columns and a rectangular shape.

    Examples
    --------
    >>> import numpy as np
    >>> from landlab import HexModelGrid
    >>> from landlab.ca.boundaries.hex_lattice_tectonicizer import LatticeNormalFault
    >>> pid = np.arange(25, dtype=int)
    >>> pdata = np.arange(25)
    >>> ns = np.arange(25, dtype=int)
    >>> grid = HexModelGrid(5, 5, 1.0, orientation='vertical', shape='rect', reorient_links=True)
    >>> lnf = LatticeNormalFault(0.0, grid, ns, pid, pdata, 0.0)
    >>> lnf.first_fw_col
    1
    >>> lnf.num_fw_rows
    array([0, 1, 3, 4, 5])
    >>> lnf.incoming_node
    array([1, 3, 4, 6])
    >>> lnf.outgoing_node
    array([12, 17, 19, 22])

    >>> pid = arange(16, dtype=int)
    >>> ns = arange(16, dtype=int)
    >>> pdata = arange(16)
    >>> grid = HexModelGrid(4, 4, 1.0, orientation='vertical', shape='rect', reorient_links=True)
    >>> lnf = LatticeNormalFault(0.0, grid, ns, pid, pdata, 0.0)
    >>> lnf.num_fw_rows
    array([0, 1, 3, 4])
    >>> lnf.incoming_node
    array([1, 2, 5])
    >>> lnf.outgoing_node
    array([ 7, 11, 15])
    >>> lnf.do_offset(rock_state=16)
    >>> ns
    array([ 0, 16, 16, 16,  4, 16,  6,  1,  8,  2, 10,  5, 12, 13, 14,  9])
    >>> lnf.propid
    array([ 0,  7, 11,  3,  4, 15,  6,  1,  8,  2, 10,  5, 12, 13, 14,  9])

    >>> pid = arange(20, dtype=int)
    >>> ns = arange(20, dtype=int)
    >>> pdata = arange(20)
    >>> grid = HexModelGrid(4, 5, 1.0, orientation='vertical', shape='rect', reorient_links=True)
    >>> lnf = LatticeNormalFault(0.0, grid, ns, pid, pdata, 0.0)
    >>> lnf.incoming_node
    array([1, 3, 4, 6])
    >>> lnf.outgoing_node
    array([12, 14, 17, 19])
    >>> lnf.do_offset(rock_state=20)
    >>> ns
    array([ 0, 20, 20, 20, 20,  5, 20, 20,  8,  1, 10,  3,  4, 13,  6, 15, 16,
            9, 18, 11])
    >>> lnf.propid
    array([ 0, 12,  2, 14, 17,  5, 19,  7,  8,  1, 10,  3,  4, 13,  6, 15, 16,
            9, 18, 11])
    """

    def __init__(self, fault_x_intercept=0.0, grid=None, node_state=None,
                 propid=None, prop_data=None, prop_reset_value=None):
        """
        Create and initialize a LatticeNormalFault object.

        Examples
        --------
        >>> import numpy as np
        >>> from landlab import HexModelGrid
        >>> from landlab.ca.boundaries.hex_lattice_tectonicizer import LatticeNormalFault
        >>> pid = np.arange(25, dtype=int)
        >>> pdata = np.arange(25)
        >>> ns = np.arange(25, dtype=int)
        >>> grid = HexModelGrid(5, 5, 1.0, orientation='vertical', shape='rect', reorient_links=True)
        >>> lnf = LatticeNormalFault(-0.01, grid, ns, pid, pdata, 0.0)
        >>> lnf.first_fw_col
        0
        >>> lnf.num_fw_rows
        array([1, 2, 4, 5, 5])
        >>> lnf.incoming_node
        array([0, 1, 3, 4, 6])
        >>> lnf.outgoing_node
        array([12, 17, 19, 22, 24])

        >>> pid = np.arange(16, dtype=int)
        >>> pdata = np.arange(16)
        >>> ns = np.arange(16, dtype=int)
        >>> grid = HexModelGrid(4, 4, 1.0, orientation='vertical', shape='rect', reorient_links=True)
        >>> lnf = LatticeNormalFault(0.0, grid, ns, pid, pdata, 0.0)
        >>> lnf.first_fw_col
        1
        >>> lnf.num_fw_rows
        array([0, 1, 3, 4])
        >>> lnf.incoming_node
        array([1, 2, 5])
        >>> lnf.outgoing_node
        array([ 7, 11, 15])

        >>> pid = np.arange(45, dtype=int)
        >>> pdata = np.arange(45)
        >>> ns = np.arange(45, dtype=int)
        >>> grid = HexModelGrid(5, 9, 1.0, orientation='vertical', shape='rect', reorient_links=True)
        >>> lnf = LatticeNormalFault(0.0, grid, ns, pid, pdata, 0.0)
        >>> lnf.first_fw_col
        1
        >>> lnf.num_fw_rows
        array([0, 1, 3, 4, 5, 5, 5, 5, 5])
        >>> lnf.incoming_node
        array([ 1,  2,  3,  5,  6,  7,  8, 10, 11, 12])
        >>> lnf.outgoing_node
        array([22, 31, 33, 34, 35, 38, 39, 40, 43, 44])
        """
        # Do the base class init
        super(LatticeNormalFault, self).__init__(grid, node_state, propid,
                                                 prop_data, prop_reset_value)
        # Set up data structures:
        #   Make sure the footwall location is such that the fault actually
        #   cuts across the grid. This means the x intercept has to be, at
        #   the very least, no smaller than the biggest x-coordinate, and if
        #   there is an even number of columns, it must be smaller than that
        #   number minus 1/tangent 60 degrees (0.57735)
        assert fault_x_intercept < (amax(self.grid.node_x) - 0.57735), 'err'

        #   Figure out which nodes are and are not within the footwall
        in_footwall = (self.grid.node_y < _TAN60 * (self.grid.node_x -
                       fault_x_intercept))

        # Helpful to have an array of node IDs for the bottom full row. Because
        # the nodes in the bottom row are staggered in a vertical, rectangular
        # hex grid, the IDs go: 0, M, 1, M+1, 2, M+2, ... etc., where M is half
        # the number of columns, rounded up (so, for example, 3 for a 5- or 6-
        # column grid, etc.)
        half_num_cols = (self.nc + 1) // 2
        bottom_row_node_id = (arange(self.nc) // 2 +
                              (arange(self.nc) % 2) * half_num_cols)

        # Also useful to remember the number of even-numbered columns
        self.n_even_cols = (self.nc + 1) // 2

        #   Find the first of the bottom-row nodes that lies in the footwall.
        # This loop exploits the fact that nodes are numbered in an order
        # sorted by x then y, and that the bottom row is staggered, with node
        # zero being "low", like: ,',', etc.
        self.first_fw_col = 0
        n = 0
        while not in_footwall[bottom_row_node_id[n]]:
            n += 1
            self.first_fw_col += 1
            assert n < self.nc, 'overflow in loop'

        #   Remember the number of footwall rows in each column
        self.num_fw_rows = zeros(self.nc, dtype=int)
        for c in range(self.nc):
            current_row = 0
            while (current_row < self.nr and
                   in_footwall[bottom_row_node_id[c] + self.nc*current_row]):
                self.num_fw_rows[c] += 1
                current_row += 1

        # If we're handling properties and property IDs, we need to do some
        # setup
        if self.propid is not None:

            # We want to remember the node IDs of two sets of nodes: those
            # whose contents will vanish off the right-hand side (and possibly
            # the top) of the grid with each offset step, and those that gain
            # new ("rock") contents with each such step.
            #
            # First, let's find the latter set, which we'll call
            # "incoming_node" because material is flowing into these nodes from
            # below. To do this, we'll go column-by-column, starting with
            # the first column that has nodes in the footwall, and going to
            # the next-to-last column. Even-numbered columns have *two*
            # incoming nodes at their base (if the footwall reaches that high),
            # whereas odd-numbered columns have only one. Note that we'll start
            # with a list (so we can append), and then convert to an array
            # that belongs to this class.
            incoming_node_list = []
            lower_right_node = (self.nc % 2) * (self.nc // 2) + \
                               ((self.nc + 1) % 2) * (self.nc - 1)
            for n in range(0, self.nc + (self.nc + 1) // 2):
                if in_footwall[n] and ((n - lower_right_node) % self.nc) != 0:
                    incoming_node_list.append(n)

            # Convert to a numpy array that belongs to the class (so we can
            # use it in the do_offset() method).
            self.incoming_node = array(incoming_node_list, dtype=int)

            # Next, find the IDs the "outgoing" nodes. There will always be
            # some of these along the right side of the grid. Depending on the
            # height of the grid and the position of the fault, there may or
            # may not also be some along the top.
            #
            # To find out which nodes will be exiting the grid, we use
            # geometry: simply find out which nodes are (1) within the footwall
            # and (2) the tectonic offset would take them beyond the grid
            # perimeter. We already have information about the first condition.
            # For the second, let's start by defining the grid edges. The
            # y coordinates of the top full row of nodes will either be NR - 1
            # (even-numbered columns) or NR - 1/2. So we'll take as a
            # reasonable boundary line NR - 1/4: any node that would move above
            # this point is definitely out of the grid. Note that the vertical
            # offset of nodes during each earthquake will be 1.5, so if we were
            # to offset a top-row node, it would move to y = NR + 1/2 or
            # y = NR + 1. Odd-numbered columns in the next-from-top row will
            # move to y = NR, which is out of bounds, so we want to flag these
            # too, and therefore need our cutoff below this. Even-numbered
            # columns in the next-from-top row will end up at y = NR - 1/2,
            # which is within the grid. So our cutoff must be between NR - 1/2
            # and NR. Hence the choise of NR - 1/4 as the effective "top" of
            # the grid.
            top_grid_edge = self.nr - 0.25

            # The right-hand edge of the grid is a simpler case, because our
            # grid is vertically oriented and there is no stagger on the right
            # and left sides. So we simply make it half a column beyond the
            # right-most column. Column width is sqrt(3), the last column is
            # NC - 1, so the right edge y coordinate is (NC - 1/2) x sqrt(3)/2
            right_grid_edge = (self.nc - 0.5) * (sqrt(3.0) / 2.0)

            # To save a repeated calculation in a loop, we'll find a threshold
            # x and y coordinate beyond which any node offset would take them
            # off the grid.
            x_threshold = right_grid_edge - (sqrt(3.0) / 2.0)
            y_threshold = top_grid_edge - 1.5

            # Actually it turns out there is a third criterion. One or two
            # nodes in the lower-right corner could be counted as both
            # "incoming" (they're at the bottom of the grid) AND outgoing
            # (they're on the right-hand side). We ignored these in setting up
            # incoming, so we should ignore them for outgoing too. This is
            # easy: any nodes on the right side (x > x_threshold) that are also
            # near the bottom (y < 1.25) should be ignored.

            # Now march through all nodes, placing those on the list that meet
            # our criteria. Yes, it's slow to do this as a Python loop, but we
            # only do it once.
            outgoing_node_list = []
            for n in range(self.grid.number_of_nodes):
                if (((self.grid.node_x[n] > x_threshold and
                      self.grid.node_y[n] > 1.25) or
                     self.grid.node_y[n] > y_threshold) and in_footwall[n]):
                    outgoing_node_list.append(n)

            # Finally, convert the outgoing node list to an array stored in
            # this object
            self.outgoing_node = array(outgoing_node_list, dtype=int)

    def do_offset(self, rock_state=1):
        """Apply 60-degree normal-fault offset.

        Offset is applied to a hexagonal grid with vertical node orientation
        and rectangular arrangement of nodes.

        Parameters
        ----------
        rock_state : int
            State code to apply to new cells introduced along bottom row.

        Examples
        --------
        >>> import numpy as np
        >>> from landlab.ca.boundaries.hex_lattice_tectonicizer import LatticeNormalFault
        >>> from landlab import HexModelGrid
        >>> pid = np.arange(25, dtype=int)
        >>> pdata = np.arange(25)
        >>> ns = np.arange(25, dtype=int)
        >>> grid = HexModelGrid(5, 5, 1.0, orientation='vertical', shape='rect', reorient_links=True)
        >>> lnf = LatticeNormalFault(0.0, grid, ns, pid, pdata, 0.0)
        >>> lnf.do_offset(rock_state=25)
        >>> ns
        array([ 0, 25, 25, 25, 25,  5, 25, 25,  8,  1, 10,  3,  4, 13,  6, 15, 16,
                9, 18, 11, 20, 21, 14, 23, 24])
        >>> lnf.propid
        array([ 0, 12,  2, 17, 19,  5, 22,  7,  8,  1, 10,  3,  4, 13,  6, 15, 16,
                9, 18, 11, 20, 21, 14, 23, 24])
        """

        # If we need to shift the property ID numbers, we'll first need to
        # record the property IDs in those nodes that are about to "shift off
        # the grid" (or rather, their contents will shift) due to tectonic
        # motion. We'll call these "nodes to replace".
        if self.propid is not None:

            # Now, remember the property IDs of the nodes along the right side
            # and possibly top that are about to shift off the grid
            propids_for_incoming_nodes = self.propid[self.outgoing_node]
            #print 'pid for new base:',propids_for_incoming_nodes

        # We go column-by-column, starting from the right side
        for c in range(self.grid.number_of_node_columns - 1,
                       self.first_fw_col - 1, -1):

            # Odd-numbered rows are shifted up in the hexagonal, vertically
            # oriented lattice
            row_offset = 2 - (c % 2)

            # Number of base nodes in the footwall in this column (1 or 2).
            n_base_nodes = min(self.num_fw_rows[c], row_offset)

            # ID of the bottom footwall node in this column
            bottom_node = (c // 2) + ((c % 2) * self.n_even_cols)

            # The bottom 1 or 2 nodes in this column are set to rock
            self.node_state[bottom_node] = rock_state
            if n_base_nodes == 2:
                self.node_state[bottom_node+self.nc] = rock_state

            # "indices" here contains the array indices of those nodes in this
            # column that are to be replaced by the ones in the column to the
            # left and down one or two nodes. We do this replacement if indices
            # contains any data.
            first_repl = bottom_node + n_base_nodes * self.nc
            last_repl = first_repl + (self.num_fw_rows[c] - \
                        (n_base_nodes + 1)) * self.nc
            indices = arange(first_repl, last_repl + 1, self.nc)

            if len(indices) > 0:
                offset = (self.nc + ((self.nc + 1) // 2) +
                          ((c + 1) % 2) * ((self.nc + 1) % 2))
                self.node_state[indices] = self.node_state[indices-offset]
                if self.propid is not None:
                    self.propid[indices] = self.propid[indices-offset]

        if self.propid is not None:
            self.propid[self.incoming_node] = propids_for_incoming_nodes
            self.prop_data[self.propid[self.incoming_node]] = self.prop_reset_value

        if self.first_fw_col==0:
            self.node_state[:self.n_footwall_rows[0]] = rock_state


class LatticeUplifter(HexLatticeTectonicizer):
    """Handles vertical uplift of interior (not edges) for a hexagonal lattice
    with vertical node orientation and rectangular node arrangement.
    """
    def __init__(self, grid=None, node_state=None, propid=None, prop_data=None, prop_reset_value=None):
        """
        Create and initialize a LatticeUplifter

        Examples
        --------
        >>> lu = LatticeUplifter()
        >>> lu.inner_base_row_nodes
        array([1, 3, 4])

        >>> hg = HexModelGrid(5, 6, 1.0, orientation='vertical', shape='rect', reorient_links=True)
        >>> lu = LatticeUplifter(grid=hg)
        >>> lu.inner_base_row_nodes
        array([1, 2, 3, 4])
        """
        # Do the base class init
        super(LatticeUplifter, self).__init__(grid, node_state, propid,
                                              prop_data, prop_reset_value)

        # Remember the IDs of nodes on the bottom row
        self.inner_base_row_nodes = zeros(self.nc - 2, dtype=int)
        n_in_lower = (self.nc // 2) - 1  # num inner nodes bottom row
        upper_start = (self.nc + 1) // 2
        n_in_upper = upper_start - 1
        self.inner_base_row_nodes[:n_in_lower] = arange(1, n_in_lower + 1)
        self.inner_base_row_nodes[n_in_lower:] = arange(upper_start,
                                                        upper_start + \
                                                        n_in_upper)

        if self.propid is not None:
            self.inner_top_row_nodes = self.inner_base_row_nodes + \
                                       ((self.nr - 1) * self.nc)


    def uplift_interior_nodes(self, rock_state=1):
        """
        Simulate 'vertical' displacement by shifting contents of node_state

        Examples
        --------
        >>> lu = LatticeUplifter()
        >>> lu.node_state[:] = arange(len(lu.node_state))
        >>> lu.uplift_interior_nodes(rock_state=25)
        >>> lu.node_state # doctest: +NORMALIZE_WHITESPACE
        array([ 0, 25,  2, 25, 25,
                5,  1,  7,  3,  4,
               10,  6, 12,  8,  9,
               15, 11, 17, 13, 14,
               20, 16, 22, 18, 19])
        """

        # Shift the node states up by a full row. A "full row" includes two
        # staggered rows.
        for r in range(self.nr - 1, 0, -1):
            # This row gets the contents of the nodes 1 row down
            self.node_state[self.inner_base_row_nodes+self.nc*r] = \
                    self.node_state[self.inner_base_row_nodes+self.nc*(r-1)]

        # Fill the bottom rows with "fresh material" (code = rock_state)
        self.node_state[self.inner_base_row_nodes] = rock_state

        # Shift the node states up by two rows: two because the grid is
        # staggered, and we don't want any horizontal offset.

        # STILL TO DO: MAKE SURE THIS HANDLES WRAP PROPERLY (I DON'T THINK
        # IT DOES NOW)
        # If propid (property ID or index) is defined, shift that too.
        if self.propid is not None:
            top_row_propid = self.propid[self.inner_top_row_nodes]
            for r in range(self.nr-1, 1, -1):
                self.propid[self.inner_base_row_nodes+self.nc*r] =  \
                            self.propid[self.inner_base_row_nodes+self.nc*(r-2)]
            self.propid[self.inner_base_row_nodes] = top_row_propid
            self.prop_data[self.propid[self.inner_base_row_nodes]] = self.prop_reset_value


if __name__=='__main__':
    import doctest
    doctest.testmod()



#! /usr/bin/env python
"""The Landlab

:Package name: TheLandlab
:Release date: 2013-03-24
:Authors: Greg Tucker, Nicole Gasparini, Erkan Istanbulluoglu, Daniel Hobley,
    Sai Nudurupati, Jordan Adams, Eric Hutton
:URL: http://csdms.colorado.edu/trac/landlab
:License: MIT
"""

from __future__ import absolute_import

__version__ = '1.0.0-beta.3'


import os
if 'DISPLAY' not in os.environ:
    try:
        import matplotlib
    except ImportError:
        import warnings
        warnings.warn('matplotlib not found', ImportWarning)
    else:
        matplotlib.use('Agg')

from .core.model_parameter_dictionary import ModelParameterDictionary
from .core.model_parameter_dictionary import (MissingKeyError,
                                              ParameterValueError)
from .core.model_component import Component
from .framework.collections import Palette, Arena, NoProvidersError
from .framework.decorators import Implements, ImplementsOrRaise
from .framework.framework import Framework
from .field.scalar_data_fields import FieldError
from .grid import *
from .plot import *

from .testing.nosetester import LandlabTester
test = LandlabTester().test
bench = LandlabTester().bench

__all__ = ['ModelParameterDictionary', 'MissingKeyError',
           'ParameterValueError', 'Component', 'Palette', 'Arena',
           'NoProvidersError', 'Implements', 'ImplementsOrRaise',
           'Framework', 'FieldError', 'LandlabTester']





import numpy as np
cimport numpy as np
cimport cython


DTYPE = np.int
ctypedef np.int_t DTYPE_t


@cython.boundscheck(False)
def _face_at_link(shape, np.ndarray[DTYPE_t, ndim=1] out):
  cdef int stride
  cdef int link
  cdef int face
  cdef int row
  cdef int col

  stride = 2 * shape[1] - 1
  n_rows = shape[0] - 1

  face = 0
  for col in range(shape[1], stride - 1):
    out[col] = face
    face += 1

  for row in range(1, n_rows):
    link = row * stride

    for col in range(0, shape[1] - 1):
      out[link + col] = face
      face += 1

    link = row * stride + shape[1] - 1
    for col in range(1, shape[1] - 1):
      out[link + col] = face
      face += 1



#! /usr/bin/env python
"""Calculate gradients of quantities over links.

Gradient calculation functions
+++++++++++++++++++++++

.. autosummary::
    :toctree: generated/

    ~landlab.grid.gradients.calc_grad_at_active_link
    ~landlab.grid.gradients.calc_grad_at_link
    ~landlab.grid.gradients.calculate_gradients_at_faces
    ~landlab.grid.gradients.calculate_diff_at_links
    ~landlab.grid.gradients.calculate_diff_at_active_links

"""

import numpy as np
from landlab.utils.decorators import use_field_name_or_array, deprecated
from landlab.core.utils import radians_to_degrees


@use_field_name_or_array('node')
def calc_grad_at_link(grid, node_values, out=None):
    """Calculate gradients of node values at links.

    Calculates the gradient in `node_values` at each link in the grid,
    returning an array of length `number_of_links`.

    Construction::

        calc_grad_at_link(grid, node_values, out=None)

    Parameters
    ----------
    grid : ModelGrid
        A ModelGrid.
    node_values : ndarray or field name (x number of nodes)
        Values at grid nodes.
    out : ndarray, optional (x number of links)
        Buffer to hold the result.

    Returns
    -------
    ndarray
        Gradients across active links.

    Examples
    --------
    >>> from landlab import RasterModelGrid
    >>> rg = RasterModelGrid(3, 4, 10.0)
    >>> z = rg.add_zeros('node', 'topographic__elevation')
    >>> z[5] = 50.0
    >>> z[6] = 36.0
    >>> calc_grad_at_link(rg, z)  # there are 17 links
    array([ 0. ,  0. ,  0. ,  0. ,  5. ,  3.6,  0. ,  5. , -1.4, -3.6,  0. ,
           -5. , -3.6,  0. ,  0. ,  0. ,  0. ])

    >>> from landlab import HexModelGrid
    >>> hg = HexModelGrid(3, 3, 10.0)
    >>> z = rg.add_zeros('node', 'topographic__elevation', noclobber=False)
    >>> z[4] = 50.0
    >>> z[5] = 36.0
    >>> calc_grad_at_link(hg, z)  # there are 11 faces
    array([ 0. ,  0. ,  0. ,  5. ,  5. ,  3.6,  3.6,  0. ,  5. , -1.4, -3.6,
            0. , -5. , -5. , -3.6, -3.6,  0. ,  0. ,  0. ])
    """
    if out is None:
        out = grid.empty(at='link')
    return np.divide(node_values[grid.node_at_link_head] -
                     node_values[grid.node_at_link_tail],
                     grid.length_of_link, out=out)


@deprecated(use='calc_grad_at_link', version='1.0beta')
def calc_grad_of_active_link(grid, node_values, out=None):
    """Calculate gradients at active links.

    Examples
    --------
    >>> import numpy as np
    >>> from landlab import RasterModelGrid
    >>> grid = RasterModelGrid((3, 4))
    >>> z = np.array([0., 0., 0., 0.,
    ...               1., 1., 1., 1.,
    ...               3., 3., 3., 3.])
    >>> grid.calc_grad_of_active_link(z)
    array([ 1.,  1.,  0.,  0.,  0.,  2.,  2.])

    This method is *deprecated*. Instead, use ``calc_grad_at_link``.

    >>> vals = grid.calc_grad_at_link(z)
    >>> vals[grid.active_links]
    array([ 1.,  1.,  0.,  0.,  0.,  2.,  2.])
    """
    return calc_grad_at_active_link(grid, node_values, out)


@deprecated(use='calc_grad_at_link', version='1.0beta')
@use_field_name_or_array('node')
def calc_grad_at_active_link(grid, node_values, out=None):
    """Calculate gradients of node values over active links.

    Calculates the gradient in *quantity* node values at each active link in
    the grid.

    Construction::

        calc_grad_at_active_link(grid, node_values, out=None)

    Parameters
    ----------
    grid : ModelGrid
        A ModelGrid.
    node_values : ndarray or field name
        Values at grid nodes.
    out : ndarray, optional
        Buffer to hold the result.

    Returns
    -------
    ndarray
        Gradients across active links.

    Examples
    --------
    >>> import numpy as np
    >>> from landlab import RasterModelGrid
    >>> grid = RasterModelGrid((3, 4))
    >>> z = np.array([0., 0., 0., 0.,
    ...               1., 1., 1., 1.,
    ...               3., 3., 3., 3.])
    >>> grid.calc_grad_at_active_link(z)
    array([ 1.,  1.,  0.,  0.,  0.,  2.,  2.])

    This method is *deprecated*. Instead, use ``calc_grad_at_link``.

    >>> vals = grid.calc_grad_at_link(z)
    >>> vals[grid.active_links]
    array([ 1.,  1.,  0.,  0.,  0.,  2.,  2.])
    """
    if out is None:
        out = grid.empty(at='active_link')
    return np.divide(node_values[grid._activelink_tonode] -
                     node_values[grid._activelink_fromnode],
                     grid.length_of_link[grid.active_links], out=out)


@deprecated(use='calc_grad_at_link', version='1.0beta')
@use_field_name_or_array('node')
def calculate_gradients_at_faces(grid, node_values, out=None):
    """Calculate gradients of node values over faces.

    Calculate and return gradient in *node_values* at each face in the grid.
    Gradients are calculated from the nodes at either end of the link that
    crosses each face.

    Construction::

        calculate_gradients_at_faces(grid, node_values, out=None)

    Parameters
    ----------
    grid : ModelGrid
        A ModelGrid.
    node_values : ndarray or field name
        Values at grid nodes.
    out : ndarray, optional
        Buffer to hold the result.

    Returns
    -------
    ndarray (x number of faces)
        Gradients across faces.

    Examples
    --------
    >>> from landlab import RasterModelGrid
    >>> rg = RasterModelGrid(3, 4, 10.0)
    >>> z = rg.add_zeros('node', 'topographic__elevation')
    >>> z[5] = 50.0
    >>> z[6] = 36.0
    >>> calculate_gradients_at_faces(rg, z)  # there are 7 faces
    array([ 5. ,  3.6,  5. , -1.4, -3.6, -5. , -3.6])

    >>> from landlab import HexModelGrid
    >>> hg = HexModelGrid(3, 3, 10.0)
    >>> z = rg.add_zeros('node', 'topographic__elevation', noclobber=False)
    >>> z[4] = 50.0
    >>> z[5] = 36.0
    >>> calculate_gradients_at_faces(hg, z)  # there are 11 faces
    array([ 5. ,  5. ,  3.6,  3.6,  5. , -1.4, -3.6, -5. , -5. , -3.6, -3.6])
    """
    if out is None:
        out = grid.empty(at='face')
    laf = grid.link_at_face
    return np.divide(node_values[grid.node_at_link_head[laf]] -
                     node_values[grid.node_at_link_tail[laf]],
                     grid.length_of_link[laf], out=out)


@use_field_name_or_array('node')
def calc_diff_at_link(grid, node_values, out=None):
    """Calculate differences of node values over links.

    Calculates the difference in quantity *node_values* at each link in the
    grid.

    Construction::

        calc_diff_at_link(grid, node_values, out=None)

    Parameters
    ----------
    grid : ModelGrid
        A ModelGrid.
    node_values : ndarray or field name
        Values at grid nodes.
    out : ndarray, optional
        Buffer to hold the result.

    Returns
    -------
    ndarray
        Differences across links.

    Examples
    --------
    >>> import numpy as np
    >>> from landlab import RasterModelGrid
    >>> rmg = RasterModelGrid((3, 3))
    >>> z = np.zeros(9)
    >>> z[4] = 1.
    >>> rmg.calc_diff_at_link(z)
    array([ 0.,  0.,  0.,  1.,  0.,  1., -1.,  0., -1.,  0.,  0.,  0.])
    """
    if out is None:
        out = grid.empty(at='link')
    node_values = np.asarray(node_values)
    return np.subtract(node_values[grid.node_at_link_head],
                       node_values[grid.node_at_link_tail], out=out)


@deprecated(use='calc_diff_at_link', version='1.0beta')
@use_field_name_or_array('node')
def calculate_diff_at_links(grid, node_values, out=None):
    """Calculate differences of node values over links.

    Examples
    --------
    >>> import numpy as np
    >>> from landlab import RasterModelGrid

    >>> grid = RasterModelGrid((3, 3))
    >>> z = np.zeros(9)
    >>> z[4] = 1.

    >>> grid.calculate_diff_at_links(z)
    array([ 0.,  0.,  0.,  1.,  0.,  1., -1.,  0., -1.,  0.,  0.,  0.])

    >>> grid.calc_diff_at_link(z)
    array([ 0.,  0.,  0.,  1.,  0.,  1., -1.,  0., -1.,  0.,  0.,  0.])
    """
    return calc_diff_at_link(grid, node_values, out)


@deprecated(use='calc_diff_at_link', version='1.0beta')
@use_field_name_or_array('node')
def calculate_diff_at_active_links(grid, node_values, out=None):
    """Calculate differences of node values over active links.

    Calculates the difference in quantity *node_values* at each active link
    in the grid.

    Construction::

        calculate_diff_at_active_links(grid, node_values, out=None)

    Parameters
    ----------
    grid : ModelGrid
        A ModelGrid.
    node_values : ndarray or field name
        Values at grid nodes.
    out : ndarray, optional
        Buffer to hold the result.

    Returns
    -------
    ndarray
        Differences across active links.
    """
    if out is None:
        out = grid.empty(at='active_link')
    node_values = np.asarray(node_values)
    return np.subtract(node_values[grid._activelink_tonode],
                       node_values[grid._activelink_fromnode], out=out)


def calc_unit_normal_at_patch(grid, elevs='topographic__elevation'):
    """Calculate and return the unit normal vector <a, b, c> to a patch.

    Parameters
    ----------
    grid : ModelGrid
        A ModelGrid.
    elevs : str or ndarray, optional
        Field name or array of node values.

    Returns
    -------
    nhat : num-patches x length-3 array
        The unit normal vector <a, b, c> to each patch.

    Examples
    --------
    >>> from landlab import HexModelGrid
    >>> mg = HexModelGrid(3, 3)
    >>> z = mg.node_x * 3. / 4.
    >>> mg.calc_unit_normal_at_patch(z)
    array([[-0.6,  0. ,  0.8],
           [-0.6,  0. ,  0.8],
           [-0.6,  0. ,  0.8],
           [-0.6,  0. ,  0.8],
           [-0.6,  0. ,  0.8],
           [-0.6,  0. ,  0.8],
           [-0.6,  0. ,  0.8],
           [-0.6,  0. ,  0.8],
           [-0.6,  0. ,  0.8],
           [-0.6,  0. ,  0.8]])
    """
    try:
        z = grid.at_node[elevs]
    except TypeError:
        z = elevs
    # conceptualize patches as sets of 3 nodes, PQR
    diff_xyz_PQ = np.empty((grid.number_of_patches, 3))
    # ^this is the vector (xQ-xP, yQ-yP, zQ-yP)
    diff_xyz_PR = np.empty((grid.number_of_patches, 3))
    P = grid.nodes_at_patch[:, 0]
    Q = grid.nodes_at_patch[:, 1]
    R = grid.nodes_at_patch[:, 2]
    x_P = grid.node_x[P]
    y_P = grid.node_y[P]
    z_P = z[P]
    diff_xyz_PQ[:, 0] = grid.node_x[Q] - x_P
    diff_xyz_PQ[:, 1] = grid.node_y[Q] - y_P
    diff_xyz_PQ[:, 2] = z[Q] - z_P
    diff_xyz_PR[:, 0] = grid.node_x[R] - x_P
    diff_xyz_PR[:, 1] = grid.node_y[R] - y_P
    diff_xyz_PR[:, 2] = z[R] - z_P
    # cross product is orthogonal to both vectors, and is the normal
    # n = <a, b, c>, where plane is ax + by + cz = d
    nhat = np.cross(diff_xyz_PQ, diff_xyz_PR)  # <a, b, c>
    nmag = np.sqrt(np.square(nhat).sum(axis=1))

    return nhat / nmag.reshape(grid.number_of_patches, 1)


def calc_slope_at_patch(grid, elevs='topographic__elevation',
                        unit_normal=None):
    """Calculate the slope (positive magnitude of gradient) at patches.

    Parameters
    ----------
    grid : ModelGrid
        A ModelGrid.
    elevs : str or ndarray, optional
        Field name or array of node values.
    unit_normal : array with shape (num_patches, 3) (optional)
        The unit normal vector to each patch, if already known.

    Returns
    -------
    slopes_at_patch : n_patches-long array
        The slope (positive gradient magnitude) of each patch.

    Examples
    --------
    >>> import numpy as np
    >>> from landlab import RasterModelGrid
    >>> mg = RasterModelGrid((4, 5))
    >>> z = mg.node_x
    >>> S = mg.calc_slope_at_patch(elevs=z)
    >>> S.size == mg.number_of_patches
    True
    >>> np.allclose(S, np.pi / 4.)
    True
    """
    if unit_normal is not None:
        assert unit_normal.shape[1] == 3
        nhat = unit_normal
    else:
        nhat = grid.calc_unit_normal_at_patch(elevs)
    dotprod = nhat[:, 2]  # by definition
    cos_slopes_at_patch = dotprod  # ...because it's now a unit vector
    slopes_at_patch = np.arccos(cos_slopes_at_patch)

    return slopes_at_patch


def calc_grad_at_patch(grid, elevs='topographic__elevation',
                       unit_normal=None, slope_magnitude=None):
    """Calculate the components of the gradient at each patch.

    Parameters
    ----------
    grid : ModelGrid
        A ModelGrid.
    elevs : str or ndarray, optional
        Field name or array of node values.
    unit_normal : array with shape (num_patches, 3) (optional)
        The unit normal vector to each patch, if already known.
    slope_magnitude : array with size num_patches (optional)
        The slope of each patch, if already known.

    Returns
    -------
    gradient_tuple : (x_component_at_patch, y_component_at_patch)
        Len-2 tuple of arrays giving components of gradient in the x and y
        directions, in the units of *units*.

    Examples
    --------
    >>> import numpy as np
    >>> from landlab import RasterModelGrid
    >>> mg = RasterModelGrid((4, 5))
    >>> z = mg.node_y
    >>> (x_grad, y_grad) = mg.calc_grad_at_patch(elevs=z)
    >>> np.allclose(y_grad, np.pi / 4.)
    True
    >>> np.allclose(x_grad, 0.)
    True
    """
    if unit_normal is not None:
        assert unit_normal.shape[1] == 3
        nhat = unit_normal
    else:
        nhat = grid.calc_unit_normal_at_patch(elevs)
    if slope_magnitude is not None:
        assert slope_magnitude.size == grid.number_of_patches
        slopes_at_patch = slope_magnitude
    else:
        slopes_at_patch = grid.calc_slope_at_patch(elevs=elevs,
                                                   unit_normal=nhat)
    theta = np.arctan2(- nhat[:, 1], - nhat[:, 0])
    x_slope_patches = np.cos(theta) * slopes_at_patch
    y_slope_patches = np.sin(theta) * slopes_at_patch

    return (x_slope_patches, y_slope_patches)


def calc_slope_at_node(grid, elevs='topographic__elevation',
                       method='patch_mean', return_components=False, **kwds):
    """Array of slopes at nodes, averaged over neighboring patches.

    Produces a value for node slope (i.e., mean gradient magnitude)
    at each node in a manner analogous to a GIS-style slope map.
    It averages the gradient on each of the
    patches surrounding the node, creating a value for node slope that
    better incorporates nonlocal elevation information. Directional
    information can still be returned through use of the return_components
    keyword.

    Note that under these definitions, it is not always true that::

        mag, cmp = mg.calc_slope_at_node(z)
        mag ** 2 == cmp[0] ** 2 + cmp[1] ** 2  # not always true

    Parameters
    ----------
    grid : ModelGrid
        A ModelGrid.
    elevs : str or ndarray, optional
        Field name or array of node values.
    method : {'patch_mean', 'Horn'}
        By equivalence to the raster version, `'patch_mean'` returns a scalar
        mean on the patches; `'Horn'` returns a vector mean on the patches.
    return_components : bool
        If True, return a tuple, (array_of_magnitude,
        (array_of_slope_x_radians, array_of_slope_y_radians)).
        If false, return an array of floats of the slope magnitude.

    Returns
    -------
    float array or length-2 tuple of float arrays
        If return_components, returns (array_of_magnitude,
        (array_of_slope_x_radians, array_of_slope_y_radians)).
        If not return_components, returns an array of slope magnitudes.

    Examples
    --------
    >>> import numpy as np
    >>> from landlab import RadialModelGrid, RasterModelGrid
    >>> mg = RasterModelGrid((4, 5), 1.)
    >>> z = mg.node_x
    >>> slopes = mg.calc_slope_at_node(elevs=z)
    >>> np.allclose(slopes, 45. / 180. * np.pi)
    True

    >>> mg = RasterModelGrid((4, 5), 1.)
    >>> z = - mg.node_y
    >>> slope_mag, cmp = mg.calc_slope_at_node(elevs=z,
    ...                                        return_components=True)
    >>> np.allclose(slope_mag, np.pi / 4.)
    True
    >>> np.allclose(cmp[0], 0.)
    True
    >>> np.allclose(cmp[1], - np.pi / 4.)
    True

    >>> mg = RadialModelGrid(num_shells=9)
    >>> z = mg.radius_at_node
    >>> slopes = mg.calc_slope_at_node(elevs=z)
    >>> mean_ring_slope = []
    >>> for i in range(10):
    ...     mean_ring_slope.append(
    ...         slopes[np.isclose(mg.radius_at_node, i)].mean())

    Notice the small amounts of numerical error here:

    >>> target_mean_ring_slope = [0.85707194785013108, 0.79363155567711452,
    ...                           0.77922185867135429, 0.78359813570962411,
    ...                           0.78433070957439543, 0.78452745144699965,
    ...                           0.78477643475446901, 0.78506472422668094,
    ...                           0.78505793680521629, 0.78661256633611021]
    >>> np.allclose(mean_ring_slope, target_mean_ring_slope)
    True
    """
    if method not in ('patch_mean', 'Horn'):
        raise ValueError('method name not understood')

    patches_at_node = np.ma.masked_where(
        grid.patches_at_node == -1, grid.patches_at_node, copy=False)

    nhat = grid.calc_unit_normal_at_patch(elevs=elevs)
    slopes_at_patch = grid.calc_slope_at_patch(elevs=elevs,
                                               unit_normal=nhat)

    # now CAREFUL - patches_at_node is MASKED
    slopes_at_node_unmasked = slopes_at_patch[patches_at_node]
    slopes_at_node_masked = np.ma.array(slopes_at_node_unmasked,
                                        mask=patches_at_node.mask)
    slope_mag = np.mean(slopes_at_node_masked, axis=1).data

    if return_components or method == 'Horn':
        (x_slope_patches, y_slope_patches) = grid.calc_grad_at_patch(
            elevs=elevs, unit_normal=nhat,
            slope_magnitude=slopes_at_patch)
        x_slope_unmasked = x_slope_patches[patches_at_node]
        x_slope_masked = np.ma.array(x_slope_unmasked,
                                     mask=patches_at_node.mask)
        x_slope = np.mean(x_slope_masked, axis=1).data
        y_slope_unmasked = y_slope_patches[patches_at_node]
        y_slope_masked = np.ma.array(y_slope_unmasked,
                                     mask=patches_at_node.mask)
        y_slope = np.mean(y_slope_masked, axis=1).data
        mean_grad_x = x_slope
        mean_grad_y = y_slope

        if method == 'Horn':
            slope_mag = np.arctan(np.sqrt(np.tan(y_slope_masked) ** 2 +
                                          np.tan(x_slope_masked) ** 2))
            return slope_mag
        else:
            return slope_mag, (mean_grad_x, mean_grad_y)

    else:
        return slope_mag


def calc_aspect_at_node(grid, slope_component_tuple=None,
                        elevs='topographic__elevation', unit='degrees'):
    """Get array of aspect of a surface.

    Calculates at returns the aspect of a surface. Aspect is returned as
    radians clockwise of north, unless input parameter units is set to
    'degrees'.

    If slope_component_tuple is provided, i.e., (slope_x, slope_y), the
    aspect will be calculated from these data.

    If it is not, it will be derived from elevation data at the nodes,
    which can either be a string referring to a grid field (default:
    'topographic__elevation'), or an nnodes-long numpy array of the
    values themselves.

    Parameters
    ----------
    grid : ModelGrid
        A ModelGrid.
    slope_component_tuple : (slope_x_array, slope_y_array) (optional)
        Tuple of components of slope in the x and y directions, defined
        on nodes, if already known. If not, provide *elevs*.
    elevs : str or array (optional)
        Node field name or node array of elevations.
        If *slope_component_tuple* is not provided, must be set, but unused
        otherwise.
    unit : {'degrees', 'radians'}
        Controls the unit that the aspect is returned as.

    Examples
    --------
    >>> from landlab import RasterModelGrid
    >>> mg = RasterModelGrid((4, 4))
    >>> z = mg.node_x ** 2 + mg.node_y ** 2
    >>> mg.calc_aspect_at_node(elevs=z)
    array([ 225.        ,  240.16585039,  255.2796318 ,  258.69006753,
            209.83414961,  225.        ,  243.54632481,  248.77808974,
            194.7203682 ,  206.45367519,  225.        ,  231.94498651,
            191.30993247,  201.22191026,  218.05501349,  225.        ])
    >>> z = z.max() - z
    >>> mg.calc_aspect_at_node(elevs=z)
    array([ 45.        ,  60.16585039,  75.2796318 ,  78.69006753,
            29.83414961,  45.        ,  63.54632481,  68.77808974,
            14.7203682 ,  26.45367519,  45.        ,  51.94498651,
            11.30993247,  21.22191026,  38.05501349,  45.        ])

    >>> mg = RasterModelGrid((4, 4), (2., 3.))
    >>> z = mg.node_x ** 2 + mg.node_y ** 2
    >>> mg.calc_aspect_at_node(elevs=z)
    array([ 236.30993247,  247.52001262,  259.97326008,  262.40535663,
            220.75264634,  234.41577266,  251.13402374,  255.29210302,
            201.54258265,  215.47930877,  235.73541937,  242.24162456,
            196.69924423,  209.43534223,  229.19345757,  236.30993247])

    Note that a small amount of asymmetry arises at the grid edges due
    to the "missing" nodes beyond the edge of the grid.
    """
    if slope_component_tuple:
        if not isinstance(slope_component_tuple, (tuple, list)):
            raise TypeError('slope_component_tuple must be tuple')
        if len(slope_component_tuple) != 2:
            raise ValueError('slope_component_tuple must be of length 2')
    else:
        try:
            elev_array = grid.at_node[elevs]
        except (KeyError, TypeError):
            assert elevs.size == grid.number_of_nodes
            elev_array = elevs

        _, slope_component_tuple = grid.calc_slope_at_node(
            elevs=elev_array, return_components=True)

    angle_from_x_ccw = np.arctan2(
        - slope_component_tuple[1], - slope_component_tuple[0])

    if unit == 'degrees':
        return radians_to_degrees(angle_from_x_ccw)
    elif unit == 'radians':
        angle_from_north_cw = (5. * np.pi / 2. -
                               angle_from_x_ccw) % (2. * np.pi)
        return angle_from_north_cw
    else:
        raise TypeError("unit must be 'degrees' or 'radians'")



#! /usr/env/python
"""Simple hexagonal Landlab cellular automaton

This file defines the HexCTS class, which is a sub-class of
CellLabCTSModel that implements a simple, non-oriented, hex-grid
CA. Like its parent class, HexCTS implements a continuous-time, stochastic,
pair-based CA. The hex grid has 3 principal directions, rather than 2 for a
raster. Hex grids are often used in CA models because of their symmetry.

Created GT Sep 2014
"""

from .celllab_cts import CellLabCTSModel
from ..grid import HexModelGrid


class HexCTS(CellLabCTSModel):
    """
    Class HexCTS implements a non-oriented hex-grid CellLab-CTS model.

    HexCTS constructor: sets number of orientations to 1 and calls
    base-class constructor.

    Parameters
    ----------
    model_grid : Landlab ModelGrid object
        Reference to the model's grid
    node_state_dict : dict
        Keys are node-state codes, values are the names associated with
        these codes
    transition_list : list of Transition objects
        List of all possible transitions in the model
    initial_node_states : array of ints (x number of nodes in grid)
        Starting values for node-state grid
    prop_data : array (x number of nodes in grid) (optional)
        Array of properties associated with each node/cell
    prop_reset_value : number or object, optional
        Default or initial value for a node/cell property (e.g., 0.0).
        Must be same type as *prop_data*.

    Examples
    --------
    >>> from landlab import HexModelGrid
    >>> from landlab.ca.celllab_cts import Transition
    >>> from landlab.ca.hex_cts import HexCTS

    >>> mg = HexModelGrid(4, 3, 1.0)
    >>> nsd = {0 : 'yes', 1 : 'no'}
    >>> xnlist = []
    >>> xnlist.append(Transition((0,1,0), (1,1,0), 1.0, 'frogging'))
    >>> nsg = mg.add_zeros('node', 'node_state_grid')
    >>> hcts = HexCTS(mg, nsd, xnlist, nsg)
    """

    def __init__(self, model_grid, node_state_dict, transition_list,
                 initial_node_states, prop_data=None, prop_reset_value=None):
        """
        HexCTS constructor: sets number of orientations to 1 and calls
        base-class constructor.

        Parameters
        ----------
        model_grid : Landlab ModelGrid object
            Reference to the model's grid
        node_state_dict : dict
            Keys are node-state codes, values are the names associated with
            these codes
        transition_list : list of Transition objects
            List of all possible transitions in the model
        initial_node_states : array of ints (x number of nodes in grid)
            Starting values for node-state grid
        prop_data : array (x number of nodes in grid) (optional)
            Array of properties associated with each node/cell
        prop_reset_value : number or object, optional
            Default or initial value for a node/cell property (e.g., 0.0).
            Must be same type as *prop_data*.
        """

        # Make sure caller has sent the right grid type
        if not isinstance(model_grid, HexModelGrid):
            raise TypeError('model_grid must be a Landlab HexModelGrid')

        # Define the number of distinct cell-pair orientations: here just 1,
        # because HexCTS represents a non-oriented CA model.
        self.number_of_orientations = 1

        # Call the LandlabCellularAutomaton.__init__() method to do the rest of
        # the initialization
        super(HexCTS, self).__init__(model_grid, node_state_dict,
                                     transition_list, initial_node_states,
                                     prop_data, prop_reset_value)


if __name__ == '__main__':
    import doctest
    doctest.testmod()
